# various utility functions used in conjunction with the FZ EIT systems
import numpy as np
import pandas as pd

import pylab as plt
import scipy.io as sio

import reda
import reda.utils.geometric_factors as geometric_factors
import reda.utils.fix_sign_with_K as fixK
import reda.importers.eit_fzj as eit_fzj


def compute_correction_factors(data, true_conductivity, elem_file, elec_file):
    """Compute correction factors for 2D rhizotron geometries, following
    Weigand and Kemna, 2017, Biogeosciences

    https://doi.org/10.5194/bg-14-921-2017

    Parameters
    ----------
    data : :py:class:`pandas.DataFrame`
        measured data
    true_conductivity : float
        Conductivity in S/m
    elem_file : string
        path to CRTomo FE mesh file (elem.dat)
    elec_file : string
        path to CRTomo FE electrode file (elec.dat)

    Returns
    -------
    correction_factors : Nx5 :py:class.`numpy.ndarray`
        measurement configurations and correction factors
        (a,b,m,n,correction_factor)
    """
    settings = {
        'rho': 100,
        'pha': 0,
        'elem': 'elem.dat',
        'elec': 'elec.dat',
        '2D': True,
        'sink_node': 100,

    }
    K = geometric_factors.compute_K_numerical(data, settings=settings)

    data = geometric_factors.apply_K(data, K)
    data = fixK.fix_sign_with_K(data)

    frequency = 100

    data_onef = data.query('frequency == {}'.format(frequency))
    rho_measured = data_onef['r'] * data_onef['k']

    rho_true = 1 / true_conductivity * 1e4
    correction_factors = rho_true / rho_measured

    collection = np.hstack((
        data_onef[['a', 'b', 'm', 'n']].values,
        np.abs(correction_factors)[:, np.newaxis]
    ))

    return collection


def apply_correction_factors(df, correction_file):
    """Apply correction factors for a pseudo-2D measurement setup. See Weigand
    and Kemna, 2017, Biogeosciences, for more information:

    https://doi.org/10.5194/bg-14-921-2017

    Parameters
    ----------
    df : :py:class:`pandas.DataFrame`
        Data container
    correction_file : string
        Path to correction file. The file must have 5 columns:
        a,b,m,n,correction_factor

    Returns
    -------

    corr_data : Nx5 :py:class:`numpy.ndarray`
        Correction files as imported from the file. Columns:
        a,b,m,n,correction_factor
    """
    if isinstance(correction_file, (list, tuple)):
        corr_data_raw = np.vstack(
            [np.loadtxt(x) for x in correction_file]
        )
    else:
        corr_data_raw = np.loadtxt(correction_file)
    A = (corr_data_raw[:, 0] / 1e4).astype(int)
    B = (corr_data_raw[:, 0] % 1e4).astype(int)
    M = (corr_data_raw[:, 1] / 1e4).astype(int)
    N = (corr_data_raw[:, 1] % 1e4).astype(int)

    corr_data = np.vstack((A, B, M, N, corr_data_raw[:, 2])).T
    corr_data[:, 0:2] = np.sort(corr_data[:, 0:2], axis=1)
    corr_data[:, 2:4] = np.sort(corr_data[:, 2:4], axis=1)

    # if 'frequency' not in df.columns:
    #     raise Exception(
    #         'No frequency data found. Are you sure this is a seit data set?'
    #     )

    gf = df.groupby(['a', 'b', 'm', 'n'])
    for key, item in gf.groups.items():
        # print('key', key)
        # print(item)
        item_norm = np.hstack((np.sort(key[0:2]), np.sort(key[2:4])))
        # print(item_norm)
        index = np.where(
            (corr_data[:, 0] == item_norm[0]) &
            (corr_data[:, 1] == item_norm[1]) &
            (corr_data[:, 2] == item_norm[2]) &
            (corr_data[:, 3] == item_norm[3])
        )[0]
        # print(index, corr_data[index])
        if len(index) == 0:
            print(key)
            # import IPython
            # IPython.embed()
            raise Exception(
                'No correction factor found for this configuration: {}'.format(
                    key
                )
            )

        factor = corr_data[index, 4]
        # apply correction factor
        for col in ('r', 'Zt', 'Vmn', 'rho_a'):
            if col in df.columns:
                df.iloc[item, df.columns.get_loc(col)] *= factor
    return corr_data


# this is data for the first test board. As far as I know nobody else has such
# an EIT system, and therefore I think it's ok to include the data here.
_resistor_data = np.array((
    (1, 4, 2, 3, 980, 10, 20),
    (2, 3, 1, 4, 980, 10, 20),
))


def check_resistor_board_measurements(data_file, reference_data_file=None,
                                      create_plot=True, **kwargs):
    """ To check basic system function a test board was built with multiple
    resistors attached to for connectors each. Measurements can thus be
    validated against known electrical (ohmic) resistances.

    Note that the normal-reciprocal difference is not yet analyzed!

    The referenc_data_file should have the following structure:
    The file contains the four-point spreads to be imported from
    the measurement. This file is a text file with four columns (A, B, M, N),
    separated by spaces or tabs. Each line denotes one measurement and its
    expected resistance, the allowed variation, and its allow difference
    towards its reciprocal counterpart: ::

        1   2   4   3   1000    1    20
        4   3   2   1   1000    1    20

    test frequency: 1Hz

    Parameters
    ----------
    data_file : string
        path to mnu0 data file
    reference_data_file: string, optional
        path to reference data file with structure as describe above. Default
        data is used if set to None
    create_plot : bool, optional
        if True, create a plot with measured and expected resistances
    kwargs : dict, optional
        kwargs will be redirected to the sEIT.import_eit_fzj call

    Returns
    -------
    fig : figure object, optional
        if create_plot is True, return a matplotlib figure
    """
    # reference_data = np.loadtxt(reference_data_file)
    # configs = reference_data[:, 0:4]
    column_names = [
        'a', 'b', 'm', 'n', 'expected_r', 'variation_r', 'variation_diffr'
    ]
    if reference_data_file is None:
        ref_data = pd.DataFrame(_resistor_data, columns=column_names)
    else:
        ref_data = pd.read_csv(
            reference_data_file,
            names=column_names,
            delim_whitespace=True,
        )
    print(ref_data)
    configs = ref_data[['a', 'b', 'm', 'n']].values.astype(int)

    seit = reda.sEIT()
    seit.import_eit_fzj(data_file, configs, **kwargs)
    seit.data = seit.data.merge(ref_data, on=('a', 'b', 'm', 'n'))

    # iterate through the test configurations
    test_frequency = 1
    failing = []
    for nr, row in enumerate(ref_data.values):
        print(nr, row)
        key = tuple(row[0:4].astype(int))
        group_abmn = seit.abmn
        if key not in group_abmn.keys:
            continue
        else:
            item = seit.abmn.get_group(key)

        expected_r = row[4]
        allowed_variation = row[5]
        # expected_r_diff = row[6]

        measured_r, measured_rdiff = item.query(
            'frequency == {}'.format(test_frequency)
        )[['r', 'rdiff']].values.squeeze()
        minr = expected_r - allowed_variation
        maxr = expected_r + allowed_variation
        if not (minr <= measured_r and maxr >= measured_r):
            print('    ', 'not passing', row)
            print('    ', minr, maxr)
            print('    ', measured_r)
            failing.append((nr, measured_r))
    if len(failing) == 0:
        failing = None
    else:
        failing = np.atleast_2d(np.array(failing))

    if create_plot:
        fig, ax = plt.subplots(1, 1, figsize=(16 / 2.54, 8 / 2.54))
        data = seit.data.query('frequency == 1')
        x = np.arange(0, data.shape[0])

        ax.plot(
            x,
            data['r'],
            '.-',
            label='data',
        )
        ax.fill_between(
            x,
            data['expected_r'] - data['variation_r'],
            data['expected_r'] + data['variation_r'],
            color='green',
            alpha=0.8,
            label='allowed limits',
        )
        if failing is not None:
            ax.scatter(
                failing[:, 0],
                failing[:, 1],
                color='r',
                label='not passing',
                s=40,
            )

        ax.legend()
        ax.set_xticks(x)
        xticklabels = [
            '{}-{} {}-{}'.format(*row) for row
            in data[['a', 'b', 'm', 'n']].values.astype(int)
        ]
        ax.set_xticklabels(xticklabels, rotation=45)

        ax.set_ylabel(r'resistance $[\Omega]$')
        ax.set_xlabel('configuration a-b m-n')
        if failing is None:
            suffix = ' PASSED'
        else:
            suffix = ''
        ax.set_title('Resistor-check for FZJ-EIT systems' + suffix)

        fig.tight_layout()
        # fig.savefig('out.pdf')
        return fig


def get_md_data_2018a(filename):
    """Return the md data of a given FZJ EIT 2018a LI calibration data file.

    This function should probably go into the importers, but for now will
    reside here until it can be properly integrated.

    Parameters
    ----------
    filename : str
        Path to eit_data.mat file generated for an LI-'calibration' run

    Returns
    -------
    md : pandas.DataFrame
        MD data

    """
    mat = sio.loadmat('eit_data.mat', squeeze_me=True)
    importer = eit_fzj.mat_version_importers['FZJ-EZ-2018A']
    md = importer._extract_md(mat, multiplexer_group=1)
    return md


def testboard_evaluation(datapath, configdat,
                         outputname, frequencies=np.logspace(-1, 4, 40),
                         error_percentage=1):
    """
    A testboard with resistors and capacitors was built to test the
    basic operation performance of eit-systems from FZJ. This function plots
    the results of measurements on this board in terms of impedance magnitude
    and phase.

    Parameters
    ----------
    datapath : str
        Path to the eit_data_mnu0.mat file containing the measurements.

    configdat: np.ndarray or txt-file
        input configuration of the used testboard configurations,
        e.g. for first two rows of the board:
        1 4 2 3
        2 3 1 4
        5 8 6 7
        6 7 5 8
        Note that normal and reciprocal measurements have to be measured.

    outputname: str
        output name of plot in png-format

    frequencies: numpy array
        frequency range (in log10-space) to compare the measurements to;
        default range is from 0.1 Hz to 10 kHz

    error_percentage: float
        percentage of allowed measurement error. The range inside this
        limit will be shown as a grey shadow in the plot.

    Returns
    -------
    fig: figure object
        Saves the plot with the given output name in the execution location of
        the script.

    """

    def calc_response(frequencies):
        # calculates theoretical |Z| and Zpha of the testboard for given
        # frequencies
        omega = 2 * np.pi * frequencies
        # settings of the specific testboard; if a new testboard with different
        # resistors/capacitors is built, parameters can be changed here
        rs = 1000
        r1 = 500

        #
        c1 = 330e-9

        r2 = 500
        c2 = 47e-6

        cp = 5e-12

        # the terms
        term1 = (r1 - 1j * omega * r1 ** 2 * c1) / \
            (1 + omega ** 2 * c1 ** 2 * r1 ** 2)
        term2 = (r2 - 1j * omega * r2 ** 2 * c2) / \
            (1 + omega ** 2 * c2 ** 2 * r2 ** 2)

        z1 = rs + term1 + term2
        z2 = - 1j / (omega * cp)

        z = 1 / (1 / z1 + 1 / z2)

        rmag = np.abs(z)
        rpha = np.arctan2(z.imag, z.real) * 1000

        return rmag, rpha

    # load configurations
    if type(configdat) == np.ndarray:
        configs = configdat
    else:
        configs = np.loadtxt(configdat)

    # load measurements
    seit = reda.sEIT()
    seit.import_eit_fzj(datapath, configs)

    # append measurements to either the "normal" or "reciprocal" list
    nor = []
    rec = []
    for i in configs:
        data = seit.abmn.get_group((i[0], i[1], i[2], i[3]))
        if (data['norrec'] == 'nor').all():
            nor.append(data)
        else:
            rec.append(data)

    # calculate theoretical testboard response and error
    rmag, rpha = calc_response(frequencies)
    error_rmag = rmag * error_percentage / 100
    error_rpha = rpha * error_percentage / 100

    # plot results
    assert len(nor) > 0 or len(rec) > 0, \
        'we got neither normal or reciprocal data'
    nr_y = max((len(nor), len(rec)))
    fig, axes = plt.subplots(
        nr_y, 2, figsize=(12, 3*len(nor)), sharex=True)

    # in case of only one measurement
    if len(nor) <= 1:
        # plot normal measurements and theoretical response
        for num, n in enumerate(nor):
            axes[0].set_title('Magnitude {} {} {} {}'.format(
                n.iloc[0]['a'], n.iloc[0]['b'],
                n.iloc[0]['m'], n.iloc[0]['n']))
            axes[0].plot(n["frequency"], n['r'],
                         marker='o', linestyle=' ', label='nor')
            axes[0].plot(frequencies, rmag, label='calculated')
            axes[0].fill_between(frequencies, rmag+error_rmag,
                                 rmag-error_rmag, color='grey', alpha=0.3)
            axes[0].set_ylabel(r'|Z| [$\Omega$]')
            axes[1].set_title('Phase {} {} {} {}'.format(
                n.iloc[0]['a'], n.iloc[0]['b'],
                n.iloc[0]['m'], n.iloc[0]['n']))
            axes[1].plot(n["frequency"], -1*n['rpha'],
                         marker='o', linestyle=' ', label='nor')
            axes[1].plot(frequencies, -1*rpha, label='calculated')
            axes[1].fill_between(
                frequencies, -1*rpha + error_rpha, -1*rpha-error_rpha,
                color='grey', alpha=0.3)
            axes[1].set_ylabel(r'-$\varphi_{Z}$ [mrad]')

        # plot reciprocal measurements
        for num, r in enumerate(rec):
            axes[0].plot(r["frequency"], r['r'],
                         marker='x', linestyle=' ', label='rec')
            axes[1].plot(r["frequency"], -1*r['rpha'],
                         marker='x', linestyle=' ', label='rec')

        # axis labels for two plots
        axes[0].set_xlabel("frequency [Hz]")
        axes[1].set_xlabel("frequency [Hz]")

    # in case of several measurements
    else:
        # plot normal measurements and theoretical response
        for num, n in enumerate(nor):
            axes[num-1][0].set_title('Magnitude {} {} {} {}'.format(
                n.iloc[0]['a'], n.iloc[0]['b'],
                n.iloc[0]['m'], n.iloc[0]['n']))
            axes[num-1][0].plot(n["frequency"], n['r'],
                                marker='o', linestyle=' ', label='nor')
            axes[num-1][0].plot(frequencies, rmag, label='calculated')
            axes[num-1][0].fill_between(frequencies, rmag+error_rmag,
                                        rmag-error_rmag, color='grey',
                                        alpha=0.3)
            axes[num-1][0].set_ylabel(r'|Z| [$\Omega$]')
            axes[num-1][1].set_title('Phase {} {} {} {}'.format(
                n.iloc[0]['a'], n.iloc[0]['b'],
                n.iloc[0]['m'], n.iloc[0]['n']))
            axes[num-1][1].plot(n["frequency"], -1*n['rpha'],
                                marker='o', linestyle=' ', label='nor')
            axes[num-1][1].plot(frequencies, -1*rpha, label='calculated')
            axes[num-1][1].fill_between(
                frequencies, -1*rpha + error_rpha, -1*rpha-error_rpha,
                color='grey', alpha=0.3)
            axes[num-1][1].set_ylabel(r'-$\varphi_{Z}$ [mrad]')

        # plot reciprocal measurements
        for num, r in enumerate(rec):
            axes[num-1][0].plot(r["frequency"], r['r'],
                                marker='x', linestyle=' ', label='rec')
            axes[num-1][1].plot(r["frequency"], -1*r['rpha'],
                                marker='x', linestyle=' ', label='rec')

        # axis labels for two bottom plots
        axes[len(nor)-1][0].set_xlabel("frequency [Hz]")
        axes[len(nor)-1][1].set_xlabel("frequency [Hz]")

    # axis scaling and legends
    for ax in axes.reshape(-1):
        ax.grid()
        ax.legend()
        ax.set_xscale("log")
        ax.set_xlim(min(frequencies), max(frequencies))

    fig.tight_layout()
    fig.savefig('{}.png'.format(outputname), dpi=300)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import tempfile
import zipfile
import pkg_resources

import reda

import matplotlib.pylab as plt


def prepare_4d_data():
    """Unzip included 4D data into a temp directory.

    Returns
    -------
    tempdir : string
        Temporary directory
    """
    outdir = tempfile.mkdtemp()
    filename = pkg_resources.resource_filename(
        'reda.testing', 'data/seit_test_data.zip'
    )
    mzip = zipfile.ZipFile(filename)
    mzip.extractall(outdir)
    return outdir

def load_data_into_container(datadir):
    seit = reda.sEIT()
    for nr in range(0, 4):
        seit.import_crtomo(
            directory=datadir + '/modV_0{}_noisy/'.format(nr),
            timestep=nr
        )
    seit.compute_K_analytical(spacing=1)
    return seit


def test_load_4d_data():
    datadir = prepare_4d_data()
    seit = load_data_into_container(datadir)


def test_check_4d_data():
    datadir = prepare_4d_data()
    seit = load_data_into_container(datadir)
    assert seit.nr_frequencies == 15
    assert seit.nr_timesteps == 4


def test_histogram_plotting():
    datadir = prepare_4d_data()
    seit = load_data_into_container(datadir)

    # frequencies into subplots
    name, figs = seit.plot_histograms('rho_a', 'frequency')
    # expect four figures, one for each timestep
    assert len(figs) == 4
    fig = figs[sorted(figs.keys())[0]]
    assert len(fig.axes) == 16
    # the last subplot should be hidden because we have only 15 frequencies
    assert fig.axes[-1].get_visible() is False
    # close the figures
    [plt.close(fig) for fig in figs.values()]

    # run the other trafos once
    name, figs = seit.plot_histograms('rho_a', 'frequency', log10=True)
    [plt.close(fig) for fig in figs.values()]
    name, figs = seit.plot_histograms('rho_a', 'frequency', lin_and_log10=True)
    [plt.close(fig) for fig in figs.values()]

    name, figs = seit.plot_histograms('rho_a', 'timestep')
    # expect 15 frequency plot, with four subplots each
    assert len(figs) == 15
    # get first figure
    fig = figs[sorted(figs.keys())[0]]
    assert len(fig.axes) == 4

    [plt.close(fig) for fig in figs.values()]
    # delete three frequency
    seit.data = seit.data.query('frequency != 0.001')
    seit.data = seit.data.query('frequency != 1.0')
    seit.data = seit.data.query('frequency != 1000.0')

    name, figs = seit.plot_histograms('rho_a', 'frequency')
    assert len(figs) == 4
    [plt.close(fig) for fig in figs.values()]

    name, figs = seit.plot_histograms('rho_a', 'timestep')
    # expect 15 frequency plot, with four subplots each
    assert len(figs) == 12
    [plt.close(fig) for fig in figs.values()]

    # take only the first timestep
    seit.data = seit.data.query('timestep == 1 and frequency == 1')
    name, figs = seit.plot_histograms('rho_a', 'timestep')



"""
columns in the containers are usually named in a simplified way to facilitate
working with them. However, as soon as we want to export data, or plot data,
meaningful and physically correct identifiers are required. This file defines
the corresponding relations.
"""
import matplotlib as mpl


rel = {
    'r': {
        'full_name': '|Z|_[Ohm]',
        'label_latex': r'$|Z|~[\Omega]$',
        'label_mpl': r'$|Z| [\Omega]$',
        'unit': r'$\Ohm$',
    },
    'rho_a': {
        'full_name': 'rhoa_[Ohm m]',
        'label_latex': r'$\rho_a$~[$\Omega$ m]',
        'label_mpl': r'$\rho_a$ [$\Omega$m]',
        'unit': r'$\Ohm m$',
    },
    'Iab': {
        'full_name': 'Iab_[mA]',
        'label_latex': r'$I_{ab}~[mA]$',
        'label_mpl': r'$I_{ab}$ [mA]',
        'unit': r'mA',
    },
    'Vmn': {
        'full_name': 'Vmn_[mV]',
        'label_latex': r'$U_{mn}~[mV]$',
        'label_mpl': r'$U_{mn}$ [mV]',
        'unit': r'mV',
    },
    'log_rho': {
        'full_name': 'log_rho_[Ohm m]',
        'label_latex': r'$\log_{10}(|\rho|~[\Omega\mbox{m}])$',
        'label_mpl': r'$\log_{10}(|\rho| [\Omega m])$',
        'unit': r'\Ohm m',
    },
    'rho': {
        'full_name': 'rho_[Ohm m]',
        'label_latex': r'$|\rho|~[\Omega\mbox{m}]$',
        'label_mpl': r'$|\rho|$ [$\Omega$m]',
        'unit': r'\Ohm m',
    },
    'phi': {
        'full_name': 'phi_[mrad]',
        'label_latex': r'$\phi~[\mbox{mrad}]$',
        'label_mpl': r'$\phi$ [mrad]',
        'unit': r'mrad',
    },
    'log_real': {
        'full_name': 'log_real_[S/m]',
        'label_latex': r"$\log_{10}(\sigma'~[\mbox{S/m}])$",
        'label_mpl': r"$\log_{10}$($\sigma$' [S/m])",
        'unit': r'S/m',
    },
    'log_imag': {
        'full_name': 'log_imag_[S/m]',
        'label_latex': r"$\log_{10}(|\sigma''|~[\mbox{S/m}])$",
        'label_mpl': r"$\log_{10}(|\sigma''| [S/m])$",
        'unit': r'S/m',
    },
}


def get_label(key):
    """Convenience function: return the label (latex version) of a given key,
    if available. Otherwise, return the key itself.
    """
    if key in rel:
        if mpl.rcParams['text.usetex']:
            return rel[key]['label_latex']
        else:
            return rel[key]['label_mpl']
    else:
        return key



# *-* coding: utf-8 *-*
import re
import itertools
import pandas as pd
import numpy as np
# from StringIO import StringIO
from io import StringIO
import os
import logging

from reda.importers.utils.decorators import enable_result_transforms


def _get_nr_of_electrodes(header_group):
    groups = itertools.groupby(
        header_group,
        lambda line: (
            line.startswith('[Number of Remote Units]')
        )
    )
    for switch, group in groups:
        if switch:
            nr_of_electrodes = int(next(group)[24:].strip()) + 1
            return nr_of_electrodes


def _parse_radic_header(header_group, dipole_mode="all"):
    """
    Parameters
    ----------
    dipole_mode: Which dipoles should be extracted
        "all"
        "before"
        "between"
        "after"
    """
    header = ''.join(header_group[1])

    nr_of_electrodes = _get_nr_of_electrodes(header.split('\n'))
    groups = itertools.groupby(
        header.split('\n'),
        lambda line: (
            line.startswith('[End Readings]') or
            line.startswith('[Begin Readings]')
        )
    )

    for key, group in groups:
        # there is only one group we are interested in: the first one
        if key:
            tmp = np.array(
                [
                    np.fromstring(
                        x, sep=' ', dtype=int
                    )
                    for x in next(groups)[1]
                ],
            )

            # curpots = tmp.T.copy()
            # curpots.resize(nr_of_electrodes + 2, tmp.shape[0])
            break

    # now we have the raw readings, pad them with zeros
    readings = []
    for raw_reading in tmp:
        voltage_rus = raw_reading[3:]
        # note that we now have the last electrode in here, too (so one more
        # than we have RUs. This electrode is always 0
        normed_reading = np.zeros(nr_of_electrodes)
        for i in voltage_rus[np.where(voltage_rus > 0)]:
            normed_reading[i - 1] = i
        readings.append(np.hstack((raw_reading[0:3], normed_reading)))

    # now generate the available configurations
    # add calibration
    reading_configs = {}
    for nr, reading in enumerate(readings):
        # nr = reading[0]
        A = reading[1]
        B = reading[2]

        voltages = []

        # loop over the RU-settings
        # we always measure to the next zero electrode
        for i in range(3, nr_of_electrodes + 3):
            for j in range(i + 1, nr_of_electrodes + 3):
                if reading[j] == 0:
                    M = i - 2
                    N = j - 2
                    all_active = ((reading[i] == 0) and (reading[j] == 0))
                    # print('a', 'b', 'm, n', A, B, M, N, all_active)
                    voltages.append((A, B, M, N, all_active))
                    break

        reading_configs[nr + 1] = np.array(voltages)
    return reading_configs


def _parse_remote_unit(ru_block):

    # add header
    text_block = next(ru_block)
    text_block = text_block.replace('Freq. /Hz', 'Freq./Hz')
    text_block = text_block.replace('K Factor/m', 'K-Factor/m')
    text_block = text_block.strip()

    text_block = re.sub(r'\s+', ' ', text_block) + os.linesep

    def prep_line(line):
        # for i in range(0, 7):
        index = 0
        for i in range(0, 5):
            index = line.find('.', index + 2)
            if index == -1 or index > len(line) - 12:
                break
            else:
                # number generally have 6 digits after the dot
                index += 6
            line = line[0: index] + ' ' + line[index:]

        # end_characters = (10, 25, 35, 45, 55)
        # for index in reversed(end_characters):
        #     line = line[0:index] + ' ' + line[index:]

        # line = ' '.join(line.split()) + r'\n'
        # import IPython
        # IPython.embed()
        line = re.sub(r'\s+', ' ', line) + os.linesep

        # calibration dot
        line = re.sub(' . ', ' nc ', line)
        line = line.lstrip()
        return line

    # now the data lines, with special attention
    for line in ru_block:
        text_block += prep_line(line)

    # text_block = ''.join(ru_block)

    tmp_file = StringIO(text_block)
    try:
        df = pd.read_csv(
            tmp_file,
            delim_whitespace=True,
            na_values=['NaN', ],
            # header=None,
            # sep=' ',
            # index_col=0,
        )
    except Exception as e:
        print(e)
        print(tmp_file.getvalue())

    df.columns = [
        'frequency',
        'rho',
        'rpha',
        'drho',
        'drpha',
        'with_calib',
        'I',
        'k',
        'time',
        'date'
    ]

    # phase is in degree, convert to mrad
    df['rpha'] *= np.pi / 180.0 * 1000

    df['r'] = df['rho'] / df['k']

    # % RA error (Ohm m)
    errorar = df['drho'] * df['rho'] / df['k']
    df['dr'] = errorar

    # % Error phase (mrad)
    try:
        errorpha = df['drpha'] * np.pi / 180.0 * 1000
    except Exception as e:
        print('Exception', e)
        print(df['dphi'])
        print(df)
        print('raw', tmp_file.getvalue())
        raise Exception('Error converting phase from degrees to mrad')

    df['drpha'] = errorpha

    # % U(V)
    voltage = df['I'] / 1000 * df['r'] / df['k']
    df['U'] = voltage

    # rename some columns
    # col_descriptions = {
    #     'rho': 'rho_[Ohm m]',
    #     'rpha': 'rpha_[mrad]',
    #     'dphi': 'dphi_[mrad]',
    #     '|Z|': '|Z|_[Ohm]',
    #     'd|Z|': 'd|Z|_[Ohm]',
    #     'U': 'U_[V]',
    # }
    # columns = df.columns.values.tolist()
    # for key in columns:
    #     if key in col_descriptions:
    #         columns[columns.index(key)] = col_descriptions[key]
    # df.columns = columns

    #
    df_sort = df.sort_index()
    return df_sort


def parse_reading(reading_block):

    groups = itertools.groupby(
        reading_block,
        lambda line: line.startswith('Remote Unit:')
    )
    index = 0
    # import IPython
    # IPython.embed()
    # logging.debug('reading groups: {0}'.format(len(groups)))

    ru_reading = []
    for key, group in groups:
        if key:
            # reading_nr = ''.join(group)[13:].strip()
            # import IPython
            # IPython.embed()
            # print(next(group).strip())
            # print(next(groups)[1])
            df_sort = _parse_remote_unit(next(groups)[1])
            # ru_reading[reading_nr] = df_sort
            ru_reading.append(df_sort)
        index += 1
    return ru_reading


def _decide_on_quadpole(config, settings):
    """

    """
    mode = settings.get('quadrupole_mode', 'after')
    logging.debug('Using quadrupole_mode: {0}'.format(mode))
    decision = True
    # print 'deciding on', config
    # we don't want duplicates
    if np.unique(config[0:4]).size != 4:
        logging.debug('FILTER failed: uniqueness {0} {1} {2} {3}'.format(
            *config[0:4]
        ))
        decision = False

    # we only want skip 3 data
    if 'filter_skip' in settings:
        if np.abs(config[3] - config[2]) != settings['filter_skip'] + 1:
            logging.debug('FILTER failed: filter_skip {0} {1} {2} {3}'.format(
                *config[0:4]
            ))
            decision = False

    if mode == 'before':
        if max(config[2:4]) > min(config[0:2]):
            logging.debug('FILTER failed: mode==before {0} {1} {2} {3}'.format(
                *config[0:4]
            ))
            decision = False
    elif mode == 'between':
        if (min(config[2:4] < min(config[0:2])) or
                max(config[2:4] > max(config[0:2]))):
            logging.debug(
                'FILTER failed: mode==between {0} {1} {2} {3}'.format(
                    *config[0:4]
                )
            )
            decision = False
    elif mode == 'after':
        if min(config[2:4]) < max(config[0:2]):
            logging.debug('FILTER failed: mode==after {0} {1} {2} {3}'.format(
                *config[0:4]
            ))
            decision = False

    return decision


def compute_quadrupoles(reading_configs, readings, settings):
    """

    """

    quadpole_data = []
    for key in sorted(readings.keys()):
        # print('key', key, len(reading_configs), type(reading_configs))
        configs_in_reading = reading_configs[key]
        reading = readings[key]
        # for configs_in_reading, reading in zip(reading_configs, readings):
        for nr, config in enumerate(configs_in_reading):
            df = reading[nr]
            df['a'] = config[0].astype(int)
            df['b'] = config[1].astype(int)
            df['m'] = config[3].astype(int)
            df['n'] = config[2].astype(int)

            # for now we only want configurations that are constructed out of
            # 'zero`d' electrodes in the "readings"-section of the config file
            if not config[4]:
                logging.debug(
                    'removing because of inactive electrodes: ' +
                    '{0} {1} {2} {3} {4}'.format(*config)
                )
                continue

            if _decide_on_quadpole(config, settings):
                quadpole_data.append(df)
    return quadpole_data


def write_crmod_file(sipdata, directory):

    if not os.path.isdir(directory):
        os.makedirs(directory)

    pwd = os.getcwd()
    os.chdir(directory)

    np.savetxt('frequencies.dat', sipdata[0].index)
    for nr, frequency in enumerate(sipdata[0].index):
        # print('f', frequency)
        filename = 'volt_{0:02}_{1}Hz.crt'.format(nr, frequency)
        with open(filename, 'w') as fid:
            fid.write('{0}\n'.format(len(sipdata)))
            for df in sipdata:
                AB = df.iloc[nr]['a'] * 1e4 + df.iloc[nr]['b']
                MN = df.iloc[nr]['m'] * 1e4 + df.iloc[nr]['n']
                line = '{0} {1} {2} {3}'.format(
                    int(AB),
                    int(MN),
                    df.iloc[nr]['r'],
                    df.iloc[nr]['rpha'],
                )

                fid.write(line + '\n')

    os.chdir(pwd)


@enable_result_transforms
def parse_radic_file(filename, settings, selection_mode="after",
                     reciprocal=None):
    """
    Import one result file as produced by the SIP256c SIP measuring device
    (Radic Research)

    Full settings dictionary: ::

        settings = {
            'filter_skip': (integer) skip dipoles we are interested in
            'quadrupole_mode': ['after'|'between'|'before'| 'all']
                               which dipoles to use from the file
        }


    Parameters
    ----------
    filename : string
        input filename, usually with the ending ".RES"
    settings : dict
        Settings for the data import, see code snippet above
    selection_mode : dict
        which voltage dipoles should be returned. Possible choices:
        "all"|"before"|"after"
    reciprocal : int|None
        If this is an integer, then assume this was a reciprocal measurement
        and the number denotes the largest RU number, N. Electrode numbers
        (a,b,m,n) will then be transformed to (N1 - a, N1 - b, N1 - m, N1 - n),
        with N1 = N + 1

    Returns
    -------
    sip_data : :py:pandas`pandas.DataFrame`
        The data contained in a data frame
    electrodes : None
        No electrode positions are imported
    topography : None
        No topography is imported

    """

    # removed : between py and pandas in line 384

    try:
        with open(filename, 'r', encoding='latin-1') as fid:
            lines = fid.readlines()
    except IOError:
        raise IOError('Radic SIP256c Datafile not found: {}'.format(filename))

    groups = itertools.groupby(
        lines,
        lambda line: line.startswith('Reading:')
    )

    # parse header
    group = next(groups)
    header_data = _parse_radic_header(group, dipole_mode='between')

    # parse readings
    reading_blocks = {}
    for key, group in groups:
        # determine reading number
        line = next(group)
        reading_nr = int(line[8: line.find('/')].strip())
        # print('reading nr', reading_nr)
        reading_blocks[reading_nr] = [x for x in next(groups)[1]]
        # print reading_blocks[reading_nr]

    # print(sorted(reading_blocks.keys()))
    # now parse the readings
    print('number of readings', len(reading_blocks))
    print('keys', sorted(reading_blocks.keys()))
    readings = {}
    for key in sorted(reading_blocks):
        # print('KEY/Reading', key)
        reading = reading_blocks[key]
        tmp = parse_reading(reading)
        readings[key] = tmp
    # print('reading keys', sorted(readings.keys()))

    logging.debug('removing calibration reading')
    # remove calibration reading
    if 0 in readings:
        del(readings[0])

    # print('readings', readings)
    sip_data_raw = compute_quadrupoles(header_data, readings, settings)

    sip_data = pd.concat(sip_data_raw)

    if reciprocal is not None and isinstance(reciprocal, int):
        sip_data['a'] = (reciprocal + 1) - sip_data['a']
        sip_data['b'] = (reciprocal + 1) - sip_data['b']
        sip_data['m'] = (reciprocal + 1) - sip_data['m']
        sip_data['n'] = (reciprocal + 1) - sip_data['n']

    return sip_data, None, None


if __name__ == '__main__':
    # filename = "20160506_03_rez_skip0.res"
    # filename = "20160506_02_rez_gradient.res"
    # filename = "t1/gradient_full.res"
    # filename = "20160506_01_rez_skip1.res"
    # filename = "20160609_04_gradient_test.res"

    # filename = "20160611_03_skip3_nor.res"
    # data = parse_radic_file(filename)
    # write_crmod_file(data, 'p1_dd_skip3_nor')

    # filename = "20160610_02_skip3_rec.res"
    # data = parse_radic_file(filename)
    # write_crmod_file(data, 'p1_dd_skip3_rec')

    filename = "20160612_02_p2_dd_sk3_nor.res"
    data = parse_radic_file(filename)
    write_crmod_file(data, 'p2_dd_skip3_nor')



import os
import logging
import warnings
import datetime

import pandas as pd
import h5py


class tsert_base(object):
    """Functionality related to the tsert file format used for both import and
    export
    """
    # the version of the file format
    version = 0.1

    def __init__(self, filename):
        """

        Parameters
        ----------
        filename : str
            filename to work on

        """
        # in case the file is opened, keep the object here
        self.fid = None
        self.fid_mode = None

        self._setup_logger()
        self.filename = filename
        if os.path.isfile(filename):
            self.check_format(filename)

    def check_format(self, filename):
        """Check if a given file is actually a tsert file with correct version
        number
        """
        self._open_file()
        assert 'file_format' in self.fid.attrs, \
            'Attribute "file_format" is missing'
        assert 'format_version' in self.fid.attrs, \
            'Attribute "format_version" is missing'

        assert self.fid.attrs['file_format'] == 'tsert'
        assert self.fid.attrs['format_version'] == self.version
        self._close_file()

    def _setup_logger(self):
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        )

        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)

        # add formatter to ch
        ch.setFormatter(formatter)

        # add ch to logger

        logger = logging.getLogger(__name__)
        print('TODO: Fix logging format')
        # while logger.hasHandlers():
        #     logger.removeHandler(logger.handlers[0])
        # logger.addHandler(ch)
        logger.setLevel(logging.INFO)

        self.logger = logger

    def _open_file(self, mode='r'):
        if self.fid is None or not self.fid or mode != self.fid_mode:
            if self.fid:
                self.fid.close()
            self.fid = h5py.File(self.filename, mode)
            self.fid_mode = mode
        return self.fid

    def _close_file(self):
        if self.fid:
            self.fid.close()
            self.fid_mode = None
            self.fid = None

    def get_index(self):
        """Return the index of the file"""
        if not os.path.isfile(self.filename):
            return self.get_empty_index()
        else:
            key = '/INDEX/index'
            self._open_file()
            if key in self.fid:
                index = pd.read_hdf(
                    self.filename,
                    key,
                )
            else:
                index = self.get_empty_index()
            self._close_file()
            return index


class tsert_export(tsert_base):
    """The TSERT file format -- export functions

    TSERT: Time-series electrical resistivity tomography
    """
    def __init__(self, filename):
        super().__init__(filename)
        if not os.path.isfile(filename):
            self.add_file_information()

    def write_index(self, data_index):
        key = '/INDEX/index'
        self._open_file('a')
        if key in self.fid:
            del self.fid[key]

        self._close_file()
        data_index.to_hdf(
            self.filename,
            key,
            append=True,
        )

    def get_empty_index(self):
        """Initialize an empty index"""
        index = pd.DataFrame(columns=['value', ])
        # index.index.name = 'index'
        return index

    def add_file_information(self):
        self._open_file('w')
        self.fid.attrs['file_format'] = 'tsert'
        self.fid.attrs['format_version'] = self.version
        self._close_file()

    def _add_to_attr(self, f, base, metadata):
        if base not in f:
            f.create_group(base)

        allowed_types = (
            str,
            int,
            float,
            datetime.datetime,
        )
        for key, item in metadata.items():
            if isinstance(item, allowed_types):
                f[base].attrs[key] = item
            elif isinstance(item, dict):
                new_base = base + '/' + key
                self._add_to_attr(f, new_base, item)

    def test_add_metadata(self):
        """Test function that investigates how metadata can be added to the hdf
        file
        """
        f = self._open_file('a')
        metadata = {
            'a': 'pla',
            'b': 'bum',
            'c': {
                'd': 1,
                'e': 1000,
            },
        }

        self._add_to_attr(f, 'METADATA', metadata)

        f.close()

    def add_metadata(self, metadata):
        """
        """
        f = self._open_file('a')
        self._add_to_attr(f, 'METADATA', metadata)
        f.close()

    def add_data(self, data, version, **kwargs):
        """Add data to the tsert file.

        Parameters
        ----------
        data : pandas.DataFrame

        """
        assert 'timestep' in data.columns, "timestep column must be present"

        self.logger.info('Exporting to tsert: {}'.format(self.filename))
        self._close_file()
        g = data.groupby('timestep')

        data_index = self.get_index()

        # we must check that the types of the new data and the old data
        # timesteps match (we allow arbitrary types of timestep keys)
        if data_index.shape[0] > 0:
            assert data['timestep'].dtype == data_index['value'].dtype, \
                'types of timestep-keys do not match: new: {} old: {}'.format(
                    data['timestep'].dtype,
                    data_index['value'].dtype
                )

        for timestep, item in g:
            print('@@@@@@@@@@@@@@@@@@@@@@@')
            print(timestep)
            print('index', data_index)
            # try to find the timestep in the index
            row = data_index.where(
                data_index['value'] == timestep
            ).dropna().index.values
            # print('row result', row, len(row))
            if len(row) == 1:
                # found one entry
                ts_key = row[0]
            elif len(row) == 0:
                # no entry
                if len(data_index.index) == 0:
                    ts_key = 0
                else:
                    ts_key = data_index.index.max() + 1
            elif len(row) > 1:
                raise Exception('We only allow unique values in the index')

            print('ts_key:', ts_key)

            key = '/'.join((
                'ERT_DATA',
                '{}'.format(ts_key),
                version,
            ))
            print('key', key)
            self._open_file('a')
            if key in self.fid:
                # delete data before adding it
                del self.fid[key]
            self._close_file()

            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                item.to_hdf(
                    self.filename,
                    key,
                    append=True,
                    # complevel=9,
                    # complib='lzo',
                )

            # update index
            data_index.loc[ts_key] = timestep
            if data_index.shape[
                    0] == 1 and data_index['value'].dtype == object:
                data_index['value'] = data_index['value'].astype(
                    type(timestep))

            # write to file
            self.write_index(data_index)

    def set_electrode_positions(self, electrode_positions):
        """Write electrode positions into the file. Existing positions will be
        deleted.

        Parameters
        ----------
        electrode_positions : pandas.DataFrame
            Electrode positions, columns x,y,z required. Do nothing if None.
        """
        if electrode_positions is None:
            return
        assert isinstance(electrode_positions, pd.DataFrame), \
            "electrode_positions must be a pandas DataFrame with cols x,y,z"
        assert 'x' in electrode_positions.columns, "column x not found"
        assert 'y' in electrode_positions.columns, "column y not found"
        assert 'z' in electrode_positions.columns, "column z not found"

        # this path in the hdf-file is hard coded according to the tsert specs
        key = '/ELECTRODES/ELECTRODE_POSITIONS'
        self.logger.info('Adding electrode positions to key: {}'.format(key))
        f = self._open_file('a')
        if key in f:
            del f[key]
        f.close()
        # make sure the file is closed
        self._close_file()
        # write to file
        electrode_positions.to_hdf(
            self.filename, key, append=True,
            # complevel=9,
            # complib='lzo',
        )

    def set_topography(self, topography):
        """Write topography into the file. Existing positions will be
        deleted.

        Parameters
        ----------
        topography : pandas.DataFrame|None
            Topography positions, columns x,y,z required. Do nothing if None.
        """
        if topography is None:
            return
        assert isinstance(topography, pd.DataFrame), \
            "electrode_positions must be a pandas DataFrame with cols x,y,z"
        assert 'x' in topography.columns, "column x not found"
        assert 'y' in topography.columns, "column y not found"
        assert 'z' in topography.columns, "column z not found"

        # this path in the hdf-file is hard coded according to the tsert specs
        key = '/TOPOGRAPHY/topography'
        self.logger.info('Adding topography to key: {}'.format(key))
        f = self._open_file('a')
        if key in f:
            del f[key]
        f.close()
        # make sure the file is closed
        self._close_file()
        # write to file
        topography.to_hdf(
            self.filename, key, append=True,
            # complevel=9,
            # complib='lzo',
        )



"""
Compute geometric factors (also referred to as K) using CRMod/CRTomo
"""
import os
import tempfile
import shutil
import subprocess

import numpy as np
import pandas as pd

from reda.utils import opt_import


import reda.utils.mpl
plt, mpl = reda.utils.mpl.setup()
CRbinaries = opt_import("crtomo.binaries")
CRcfg = opt_import("crtomo.cfg")


def _write_config_file(filename, dataframe):
    if isinstance(dataframe, pd.DataFrame):
        AB = dataframe['a'].values * 1e4 + dataframe['b'].values
        MN = dataframe['m'].values * 1e4 + dataframe['n'].values
    else:
        AB = dataframe[:, 0] * 1e4 + dataframe[:, 1]
        MN = dataframe[:, 2] * 1e4 + dataframe[:, 3]

    with open(filename, 'wb') as fid:
        fid.write('{0}\n'.format(AB.shape[0]).encode('utf-8'))
        np.savetxt(
            fid,
            np.vstack((AB, MN)).T,
            fmt='%i %i',
        )
    return np.vstack((AB, MN)).T


def _write_crmod_file(filename):
    """Write a valid crmod configuration file to filename.

    TODO: Modify configuration according to, e.g., 2D
    """
    crmod_lines = [
        '***FILES***',
        '../grid/elem.dat',
        '../grid/elec.dat',
        '../rho/rho.dat',
        '../config/config.dat',
        'F                ! potentials ?',
        '../mod/pot/pot.dat',
        'T                ! measurements ?',
        '../mod/volt.dat',
        'F               ! sensitivities ?',
        '../mod/sens/sens.dat',
        'F                ! another dataset ?',
        '1                ! 2D (=0) or 2.5D (=1)',
        'F                ! fictitious sink ?',
        '1660             ! fictitious sink node number',
        'F                ! boundary values ?',
        'boundary.dat',
    ]

    with open(filename, 'w') as fid:
        [fid.write(line + '\n') for line in crmod_lines]


def get_default_settings():
    return {
        'rho': 100,
        'elem': 'elem.dat',
        'elec': 'elec.dat',
        'sink_node': '100',
        '2D': False,
    }


def compute_K(dataframe, settings, keep_dir=False):
    """
    Parameters
    ----------
    dataframe : pandas.DataFrame
        dataframe that contains the data
    settings : dict
        with required settings, see below
    keep_dir : path
        if not None, copy modeling dir here


    settings = {
        'rho': 100,  # resistivity to use for homogeneous model, [Ohm m]
        'elem'
        'elec'
        '2D' : True|False
        'sink_node': '100',
    }

    """
    if settings is None:
        print('using default settings')
        settings = get_default_settings()

    if not os.path.isfile(settings['elem']):
        raise IOError(
            'elem file not found: {0}'.format(settings['elem'])
        )

    if not os.path.isfile(settings['elec']):
        raise IOError(
            'elec file not found: {0}'.format(settings['elec'])
        )

    # read grid file and determine nr of cells
    with open(settings['elem'], 'r') as fid:
        fid.readline()
        cell_type, cell_number, edge_number = np.fromstring(
            fid.readline().strip(),
            sep=' ',
            dtype=int,
        )

    # generate forward model as a string
    forward_model = '{0}\n'.format(cell_number)
    forward_model += '{0} {1}\n'.format(settings['rho'], 0) * cell_number

    full_path_elem = os.path.abspath(settings['elem'])
    full_path_elec = os.path.abspath(settings['elec'])

    pwd = os.getcwd()
    with tempfile.TemporaryDirectory() as invdir:
        os.chdir(invdir)
        # create tomodir directory structure
        for dir in [
            'exe',
            'mod',
            'config',
            'inv',
            'grid',
            'rho',
        ]:
            os.makedirs(dir)

        # save forward model
        with open('rho/rho.dat', 'w') as fid:
            fid.write(forward_model)

        shutil.copy(full_path_elem, 'grid/elem.dat')
        shutil.copy(full_path_elec, 'grid/elec.dat')

        print('SETTINGS')
        print(settings)

        cfg = CRcfg.crmod_config()
        if settings.get('2D', False):
            # activate 2D mode
            print('2D modeling')
            cfg['2D'] = '0'
            cfg['fictitious_sink'] = 'T'
            cfg['sink_node'] = settings.get('sink_node')
        else:
            cfg['2D'] = 1

        cfg.write_to_file('exe/crmod.cfg')

        # print crmod.cfg for information
        subprocess.call('cat exe/crmod.cfg', shell=True)

        config_orig = _write_config_file('config/config.dat', dataframe)

        os.chdir('exe')
        binary = CRbinaries.get('CRMod')
        output = subprocess.check_output(
            binary, shell=True, stderr=subprocess.STDOUT
        )
        # I couldn't bring the subprocess-function to recognize a crashed CRMod
        # call, so do it the hard way and parse the output
        return_value = int(
            output[output.find('STOP'.encode('utf-8')) + 4:].strip()
        )
        os.chdir('..')

        if return_value != 0:
            print(output)
            print('ERROR: There was an error with the call to CRMod')
            print('The crashed tomodir can be found here: {}'.format(invdir))
            exit()

        # read in results
        modeled_resistances = np.loadtxt(
            'mod/volt.dat',
            skiprows=1,
        )

        # now we have to make sure CRMod didn't change the signs
        changed_sign = (config_orig[:, 1] == modeled_resistances[:, 1])
        modeled_resistances[~changed_sign, 2] *= -1

        if settings.get('norm_factor', None) is not None:
            modeled_resistances[:, 2] /= settings.get('norm_factor')

        K = settings['rho'] / modeled_resistances[:, 2]
        if isinstance(dataframe, pd.DataFrame):
            dataframe['k'] = K
        if keep_dir is not None and not os.path.isdir(keep_dir):
            shutil.copytree('.', keep_dir)
            print('Copy of modeling dir stored here: {}'.format(keep_dir))

    os.chdir(pwd)
    return K



# -*- coding: utf-8 -*-
from io import StringIO

import pandas as pd
import numpy as np

from reda.containers.ERT import ERT
from reda.importers.utils.decorators import enable_result_transforms


@enable_result_transforms
def _parse_wenner_file(filename, settings):
    """Parse a Geotom .wen (Wenner configuration) file

    Parsing problems
    ----------------

    Due to column overflows it is necessary to make sure that spaces are
    present around the ; character. Example:

        8.000    14.000 10835948.70;   0.001  -123.1853  -1.0  23.10.2014

    """
    # read data
    with open(filename, 'r') as fid2:
        geotom_data_orig = fid2.read()

    # replace all ';' by ' ; '
    geotom_data = geotom_data_orig.replace(';', ' ; ')

    fid = StringIO()
    fid.write(geotom_data)
    fid.seek(0)

    header = [fid.readline() for i in range(0, 16)]
    header

    df = pd.read_csv(
        fid,
        delim_whitespace=True,
        header=None,
        names=(
            'elec1_wenner',
            'a_w',
            'rho_a',
            'c4',
            'c5',
            'c6',
            'c6',
            'c7',
            'c8',
            'c9',
        ),
    )

    # compute geometric factor using the Wenner formula
    df['k'] = 2 * np.pi * df['a_w']
    df['r'] = df['rho_a'] / df['k']

    Am = df['elec1_wenner']
    Bm = df['elec1_wenner'] + df['a_w']
    Mm = df['elec1_wenner'] + 3 * df['a_w']
    Nm = df['elec1_wenner'] + 2 * df['a_w']

    df['a'] = Am / 2.0 + 1
    df['b'] = Bm / 2.0 + 1
    df['m'] = Mm / 2.0 + 1
    df['n'] = Nm / 2.0 + 1

    # remove any nan values
    df.dropna(axis=0, subset=['a', 'b', 'm', 'n', 'r'], inplace=True)

    return df


def add_file(filename, settings, container=None):
    """

    Parameters
    ----------
    filename: string
        path
    settings: dict
        ::

            {
                timestep: [int|datetime], timestep relating to this measurement
            }

    container: :class:`reda.container.ERT.ERT`
        ERT container to add dataset to

    """
    timestep = settings.get('timestep', 0)

    # Wenner
    if filename.endswith('.wen'):
        data = _parse_wenner_file(filename, settings)
        # add timestep column
        data['timestep'] = timestep
    else:
        raise Exception('Not a Wenner file')

    if container is None:
        container = ERT(data)
    else:
        container.df = pd.concat((container.df, data))

    return container



"""Research Center Jülich - EIT40 system importer (2010 version)
"""
import datetime

import numpy as np
import pandas as pd


def _average_swapped_current_injections(df):
    AB = df[['a', 'b']].values

    # get unique injections
    abu = np.unique(
        AB.flatten().view(AB.dtype.descr * 2)
    ).view(AB.dtype).reshape(-1, 2)
    # find swapped pairs
    pairs = []
    alone = []
    abul = [x.tolist() for x in abu]
    for ab in abul:
        swap = list(reversed(ab))
        if swap in abul:
            pair = (ab, swap)
            pair_r = (swap, ab)
            if pair not in pairs and pair_r not in pairs:
                pairs.append(pair)
        else:
            alone.append(ab)

    # check that all pairs got assigned
    if len(pairs) * 2 + len(alone) != len(abul):
        print('len(pairs) * 2 == {0}'.format(len(pairs) * 2))
        print(len(abul))
        raise Exception(
            'total numbers of unswapped-swapped matching do not match!'
        )
    if len(pairs) > 0 and len(alone) > 0:
        print(
            'WARNING: Found both swapped configurations and non-swapped ones!'
        )

    delete_slices = []

    # these are the columns that we work on (and that are retained)
    columns = [
        'frequency', 'a', 'b', 'p',
        'Z1', 'Z2', 'Z3',
        'Il1', 'Il2', 'Il3',
        'Is1', 'Is2', 'Is3',
        'Zg1', 'Zg2', 'Zg3',
        'datetime',
    ]
    dtypes = {col: df.dtypes[col] for col in columns}

    X = df[columns].values

    for pair in pairs:
        index_a = np.where(
            (X[:, 1] == pair[0][0]) & (X[:, 2] == pair[0][1])
        )[0]
        index_b = np.where(
            (X[:, 1] == pair[1][0]) & (X[:, 2] == pair[1][1])
        )[0]
        # normal injection
        A = X[index_a, :]
        # swapped injection
        B = X[index_b, :]
        # make sure we have the same ordering in P, frequency
        diff = A[:, [0, 3]] - B[:, [0, 3]]
        if not np.all(diff) == 0:
            raise Exception('Wrong ordering')
        # compute the averages in A
        # the minus stems from the swapped current electrodes
        X[index_a, 4:10] = (A[:, 4:10] - B[:, 4:10]) / 2.0
        X[index_a, 10:16] = (A[:, 10:16] + B[:, 10:16]) / 2.0

        # delete the second pair
        delete_slices.append(
            index_b
        )
    if len(delete_slices) == 0:
        X_clean = X
    else:
        X_clean = np.delete(X, np.hstack(delete_slices), axis=0)

    df_clean = pd.DataFrame(X_clean, columns=columns)
    # for col in columns:
    #   # df_clean[col] = df_clean[col].astype(dtypes[col])
    df_clean = df_clean.astype(dtype=dtypes)
    return df_clean


def _extract_md(mat, **kwargs):
    return None


def _extract_emd(mat, **kwargs):
    emd = mat['EMD'].squeeze()
    # Labview epoch
    epoch = datetime.datetime(1904, 1, 1)

    def convert_epoch(x):
        timestamp = epoch + datetime.timedelta(seconds=x.astype(float))
        return timestamp

    dfl = []
    # loop over frequencies
    for f_id in range(0, emd.size):
        # print('Frequency: ', emd[f_id]['fm'])
        fdata = emd[f_id]
        # fdata_md = md[f_id]

        timestamp = np.atleast_2d(
            [convert_epoch(x) for x in fdata['Time'].squeeze()]
        ).T
        # import IPython
        # IPython.embed()
        df = pd.DataFrame(
            np.hstack((
                timestamp,
                fdata['ni'],
                fdata['nu'][:, np.newaxis],
                fdata['Z3'],
                fdata['Is3'],
                fdata['Il3'],
                fdata['Zg3'],
            )),
        )
        df.columns = (
            'datetime',
            'a',
            'b',
            'p',
            'Z1',
            'Z2',
            'Z3',
            'Is1',
            'Is2',
            'Is3',
            'Il1',
            'Il2',
            'Il3',
            'Zg1',
            'Zg2',
            'Zg3',
        )

        df['frequency'] = np.ones(df.shape[0]) * fdata['fm']

        # cast to correct type
        df['datetime'] = pd.to_datetime(df['datetime'])
        df['a'] = df['a'].astype(int)
        df['b'] = df['b'].astype(int)
        df['p'] = df['p'].astype(int)

        df['Z1'] = df['Z1'].astype(complex)
        df['Z2'] = df['Z2'].astype(complex)
        df['Z3'] = df['Z3'].astype(complex)

        df['Zg1'] = df['Zg1'].astype(complex)
        df['Zg2'] = df['Zg2'].astype(complex)
        df['Zg3'] = df['Zg3'].astype(complex)

        df['Is1'] = df['Is1'].astype(complex)
        df['Is2'] = df['Is2'].astype(complex)
        df['Is3'] = df['Is3'].astype(complex)

        df['Il1'] = df['Il1'].astype(complex)
        df['Il2'] = df['Il2'].astype(complex)
        df['Il3'] = df['Il3'].astype(complex)

        dfl.append(df)

    if len(dfl) == 0:
        return None
    df = pd.concat(dfl)

    # average swapped current injections here!
    df = _average_swapped_current_injections(df)

    # sort current injections
    condition = df['a'] > df['b']
    df.loc[condition, ['a', 'b']] = df.loc[condition, ['b', 'a']].values
    # for some reason we lose the integer casting of a and b here
    df['a'] = df['a'].astype(int)
    df['b'] = df['b'].astype(int)
    # change sign because we changed A and B
    df.loc[condition, ['Z1', 'Z2', 'Z3']] *= -1

    # average of Z1-Z3
    df['Zt'] = np.mean(df[['Z1', 'Z2', 'Z3']].values, axis=1)
    # we need to keep the sign of the real part
    sign_re = np.real(df['Zt']) / np.abs(np.real(df['Zt']))
    df['r'] = np.abs(df['Zt']) * sign_re
    # df['Zt_std'] = np.std(df[['Z1', 'Z2', 'Z3']].values, axis=1)

    df['Is'] = np.mean(df[['Is1', 'Is2', 'Is3']].values, axis=1)
    df['Il'] = np.mean(df[['Il1', 'Il2', 'Il3']].values, axis=1)
    df['Zg'] = np.mean(df[['Zg1', 'Zg2', 'Zg3']].values, axis=1)

    # "standard" injected current, in [mA]
    df['Iab'] = np.abs(df['Is']) * 1e3
    df['Iab'] = df['Iab'].astype(float)
    # df['Is_std'] = np.std(df[['Is1', 'Is2', 'Is3']].values, axis=1)
    # take absolute value and convert to mA
    df['Ileakage'] = np.abs(df['Il']) * 1e3
    df['Ileakage'] = df['Ileakage'].astype(float)

    return df



# *-* coding: utf-8 *-*
"""Read binary data from the IRIS Instruments Syscal Pro system

"""
import struct
from io import StringIO
import logging

import pandas as pd
import numpy as np

import reda
from reda.importers.utils.decorators import enable_result_transforms

logger = logging.getLogger(__name__)


@enable_result_transforms
def import_txt(filename, **kwargs):
    """Import Syscal measurements from a text file, exported as 'Spreadsheet'.

    At this point we expect only x-coordinates to be exported. As such the
    columns Spa.1 to Spa.4 are assumed to be x coordinates for a,b,m,n.

    Parameters
    ----------
    filename: str
        Input filename
    assume_regular_electrodes_x : None|tuple(nr_electrodes, spacing)
        If not None, then assume measurements were taken using a profile of
        regular electrodes in x directions. Fill in any electrodes not used in
        the data to align the logical electrode numbers wit the physical
        electrode numbers. Use the electrode spacing set in the system (usually
        a default of 1 m is used), and then change this to the real spacing
        using the 'elecs_transform_reg_spacing_x' parameter.
    elecs_transform_reg_spacing_x : tuple(old, new)|None, optional
        If not None, then assume a regular electrode spacing in x direction
        with spacing 'old'. This is often 1 m, a default of the systems.
        However, in reality often other spacings are used, and this parameter
        will lead to a transformation of the old (device-specific) spacing to
        the true spacing.
    shift_by_xyz : tuple|list|numpy.ndarray of size 1 or 2 or 3, optional
        If set, shift electrode positions by adding this vector Length of 1
        assumes that only the x coordinate of the vector differs from zero.
        Length of 2 assume a shift in (x,z) direction.

        This parameter is evaluated after the 'elecs_transform_reg_spacing_x'
        parameter - as such you can and must use real spacings in this case.
    reciprocals: int, optional
        if provided, then assume that this is a reciprocal measurements where
        only the electrode cables were switched. The provided number N is
        treated as the maximum electrode number, and denotations are renamed
        according to the equation :math:`X_n = N - (X_a - 1)`

    Returns
    -------
    data: :py:class:`pandas.DataFrame`
        Contains the measurement data
    elec_mgr : :py:class:`reda.utils.electrode_manager.electrode_manager`
        Electrode manager that manages the electrode positions
    topography: None
        No topography information is contained in the text files, so we always
        return None

    """
    # read in text file into a buffer
    with open(filename, 'r') as fid:
        text = fid.read()
    strings_to_replace = {
        'Mixed / non conventional': 'Mixed/non-conventional',
        'Date': 'Date Time AM-PM',
    }
    for key in strings_to_replace.keys():
        text = text.replace(key, strings_to_replace[key])

    buffer = StringIO(text)

    # read data file
    data_raw = pd.read_csv(
        buffer,
        # sep='\t',
        delim_whitespace=True,
    )

    # clean up column names
    data_raw.columns = [x.strip() for x in data_raw.columns.tolist()]

    # convert coordinates to logical electrode numbers
    elec_mgr = reda.electrode_manager()
    elec_mgr.set_ordering_to_sort_zyx()

    elec_mgr.add_by_position(data_raw[['Spa.1', ]].values)
    elec_mgr.add_by_position(data_raw[['Spa.2', ]].values)
    elec_mgr.add_by_position(data_raw[['Spa.3', ]].values)
    elec_mgr.add_by_position(data_raw[['Spa.4', ]].values)

    assume_regular_electrodes_x = kwargs.get(
        'assume_regular_electrodes_x', None)
    if assume_regular_electrodes_x is not None:
        nr_electrodes = assume_regular_electrodes_x[0]
        spacing = assume_regular_electrodes_x[1]
        assert isinstance(nr_electrodes, int)
        assert isinstance(spacing, float)
        elec_mgr.conform_to_regular_x_spacing(spacing, nr_electrodes)

    data = pd.DataFrame()
    data['a'] = elec_mgr.get_electrode_numbers_for_positions(
        np.atleast_2d(data_raw['Spa.1'].values).T
    )
    data['b'] = elec_mgr.get_electrode_numbers_for_positions(
        np.atleast_2d(data_raw['Spa.2'].values).T
    )
    data['m'] = elec_mgr.get_electrode_numbers_for_positions(
        np.atleast_2d(data_raw['Spa.3'].values).T
    )
    data['n'] = elec_mgr.get_electrode_numbers_for_positions(
        np.atleast_2d(data_raw['Spa.4'].values).T
    )

    transform_regular_spacing_x = kwargs.get(
        'elecs_transform_reg_spacing_x', None)
    if transform_regular_spacing_x is not None:
        assert len(transform_regular_spacing_x) == 2, \
            'spacing transform must be of length 2'
        old_spacing = transform_regular_spacing_x[0]
        new_spacing = transform_regular_spacing_x[1]
        logger.debug('transform_regular_spacing_x from {} to {}')
        elec_mgr._electrode_positions['x'] = elec_mgr._electrode_positions[
            'x'
        ] / old_spacing * new_spacing

    shift_by_xyz = kwargs.get('shift_by_xyz', None)
    if shift_by_xyz is not None:
        elec_mgr.shift_positions_xyz(shift_by_xyz)

    # [mV] / [mA]
    data['r'] = data_raw['Vp'] / data_raw['In']
    data['Vmn'] = data_raw['Vp']
    data['Iab'] = data_raw['In']

    # rename electrode denotations
    rec_max = kwargs.get('reciprocals', None)
    if rec_max is not None:
        logger.warning('renumbering electrode numbers')
        data[['a', 'b', 'm', 'n']] = rec_max + 1 - data[['a', 'b', 'm', 'n']]

    return data, elec_mgr, None


@enable_result_transforms
def import_bin(filename,  **kwargs):
    """Read a .bin file generated by the IRIS Instruments Syscal Systems
    and return a curated Pandas dataframe for further processing. This
    dataframe contains only information currently deemed important. Use the
    function reda.importers.iris_syscal_pro_binary._import_bin to extract ALL
    information from a given .bin file.

    Note that the .bin files contain electrodes positions for the electrodes.
    Electrode numbers are then inferred from those.

    Parameters
    ----------
    filename : string
        path to input filename
    assume_regular_electrodes_x : None|tuple(nr_electrodes, spacing)
        If not None, then assume measurements were taken using a profile of
        regular electrodes in x directions. Fill in any electrodes not used in
        the data to align the logical electrode numbers wit the physical
        electrode numbers. Use the electrode spacing set in the system (usually
        a default of 1 m is used), and then change this to the real spacing
        using the 'elecs_transform_reg_spacing_x' parameter.
    elecs_transform_reg_spacing_x : tuple(old, new)|None, optional
        If not None, then assume a regular electrode spacing in x direction
        with spacing 'old'. This is often 1 m, a default of the systems.
        However, in reality often other spacings are used, and this parameter
        will lead to a transformation of the old (device-specific) spacing to
        the true spacing.
    shift_by_xyz : tuple|list|numpy.ndarray of size 1 or 2 or 3, optional
        If set, shift electrode positions by adding this vector Length of 1
        assumes that only the x coordinate of the vector differs from zero.
        Length of 2 assume a shift in (x,z) direction.

        This parameter is evaluated after the 'elecs_transform_reg_spacing_x'
        parameter - as such you can and must use real spacings in this case.
    reciprocals : int, optional
        if provided, then assume that this is a reciprocal measurement where
        only the electrode cables were switched. The provided number N is
        treated as the maximum electrode number, and denotations are renamed
        according to the equation :math:`X_n = N - (X_a - 1)`
    check_meas_nums : bool, optional
        if True, then check that the measurement numbers are consecutive. Don't
        return data after a jump to smaller measurement numbers (this usually
        indicates that more data points were downloaded than are part of a
        specific measurement. Default: True
    skip_rows : int, optional
        Ignore this number of rows at the beginning, e.g., because they were
        inadvertently imported from an earlier measurement. Default: 0

    Returns
    -------
    data : :py:class:`pandas.DataFrame`
        Contains the measurement data
    elec_mgr : :py:class:`reda.utils.electrode_manager.electrode_manager`
        Electrode manager that manages the electrode positions
    topography : None
        No topography information is contained in the text files, so we always
        return None
    """

    metadata, data_raw = _import_bin(filename)

    skip_rows = kwargs.get('skip_rows', 0)
    if skip_rows > 0:
        data_raw.drop(data_raw.index[range(0, skip_rows)], inplace=True)
        data_raw = data_raw.reset_index()

    if kwargs.get(
            'check_meas_nums', True) and 'measurement_num' in data_raw.columns:
        # check that first number is 0
        if data_raw['measurement_num'].iloc[0] != 0:
            logger.warning(
                'WARNING: Measurement numbers do not start with 0 ' +
                '(did you download ALL data?)'
            )

        # check that all measurement numbers increase by one
        if not np.all(np.diff(data_raw['measurement_num'])) == 1:
            logger.warning(' '.join((
                'WARNING: ',
                'Measurement numbers are not consecutive.',
                'Perhaps the first measurement belongs to another'
                ' measurement?',
                ' Use the skip_rows parameter to skip those measurements'
            )))

        # now check if there is a jump in measurement numbers somewhere
        # ignore first entry as this will always be nan
        diff = data_raw['measurement_num'].diff()[1:]
        jump = np.where(diff != 1)[0]
        if len(jump) > 0 and not np.all(data_raw['measurement_num'] == 0):
            logger.warning(
                'WARNING: One or more jumps in measurement numbers detected')
            logger.warning('The jump indices are:')
            for jump_nr in jump:
                logger.warning(jump_nr)

            logger.info('Removing data points subsequent to the first jump')
            data_raw = data_raw.iloc[0:jump[0] + 1, :]

    if data_raw.shape[0] == 0:
        # no data present, return a bare DataFrame
        return pd.DataFrame(columns=['a', 'b', 'm', 'n', 'r']), None, None

    # convert coordinates to logical electrode numbers
    elec_mgr = reda.electrode_manager()
    elec_mgr.set_ordering_to_sort_zyx()

    elec_mgr.add_by_position(data_raw[['x_a', 'y_a', 'z_a']].values)
    elec_mgr.add_by_position(data_raw[['x_b', 'y_b', 'z_b']].values)
    elec_mgr.add_by_position(data_raw[['x_m', 'y_m', 'z_m']].values)
    elec_mgr.add_by_position(data_raw[['x_n', 'y_n', 'z_n']].values)

    assume_regular_electrodes_x = kwargs.get(
        'assume_regular_electrodes_x', None)
    if assume_regular_electrodes_x is not None:
        nr_electrodes = assume_regular_electrodes_x[0]
        spacing = assume_regular_electrodes_x[1]
        assert isinstance(nr_electrodes, int)
        assert isinstance(spacing, float)
        elec_mgr.conform_to_regular_x_spacing(spacing, nr_electrodes)

    data = pd.DataFrame()
    data['a'] = elec_mgr.get_electrode_numbers_for_positions(
        data_raw[['x_a', 'y_a', 'z_a']].values
    )
    data['b'] = elec_mgr.get_electrode_numbers_for_positions(
        data_raw[['x_b', 'y_b', 'z_b']].values
    )
    data['m'] = elec_mgr.get_electrode_numbers_for_positions(
        data_raw[['x_m', 'y_m', 'z_m']].values
    )
    data['n'] = elec_mgr.get_electrode_numbers_for_positions(
        data_raw[['x_n', 'y_n', 'z_n']].values
    )

    transform_regular_spacing_x = kwargs.get(
        'elecs_transform_reg_spacing_x', None)
    if transform_regular_spacing_x is not None:
        assert len(transform_regular_spacing_x) == 2, \
            'spacing transform must be of length 2'
        old_spacing = transform_regular_spacing_x[0]
        new_spacing = transform_regular_spacing_x[1]
        logger.debug('transform_regular_spacing_x from {} to {}')
        elec_mgr._electrode_positions['x'] = elec_mgr._electrode_positions[
            'x'
        ] / old_spacing * new_spacing

    shift_by_xyz = kwargs.get('shift_by_xyz', None)
    if shift_by_xyz is not None:
        elec_mgr.shift_positions_xyz(shift_by_xyz)

    # [mV] / [mA]
    data['r'] = data_raw['vp'] / data_raw['Iab']
    data['Vmn'] = data_raw['vp']
    data['Iab'] = data_raw['Iab']

    if 'vab' in data_raw.columns:
        data['vab'] = data_raw['vab']

    data['mdelay'] = data_raw['mdelay']
    data['Tm'] = data_raw['Tm']
    data['Mx'] = data_raw['Mx']
    data['chargeability'] = data_raw['m']
    data['q'] = data_raw['q']

    # rename electrode denotations
    rec_max = kwargs.get('reciprocals', None)
    if rec_max is not None:
        logger.info(
            'renumbering electrode numbers for reciprocal measurements')
        data[['a', 'b', 'm', 'n']] = rec_max + 1 - data[['a', 'b', 'm', 'n']]

    return data, elec_mgr, None


def _import_bin(filename):
    """Read a .bin file generated by the IRIS Instruments Syscal Pro System

    Parameters
    ----------
    filename : string
        Path to input filename

    Returns
    -------
    metadata : dict
        General information on the measurement
    df : :py:class:`pandas.DataFrame`
        dataframe containing all measurement data

    """
    fid = open(filename, 'rb')

    def fget(fid, fmt, tsize):
        buffer = fid.read(tsize)
        result_raw = struct.unpack(fmt, buffer)
        if len(result_raw) == 1:
            return result_raw[0]
        else:
            return result_raw

    # determine overall size
    fid.seek(0, 2)
    total_size = fid.tell()
    # print('total size', total_size)

    # start from the beginning
    fid.seek(0)

    # read version
    buffer = fid.read(4)
    version = struct.unpack('I', buffer)
    # print('version', version)
    buffer = fid.read(1)
    typedesyscal = struct.unpack('c', buffer)[0]
    syscal_type = int.from_bytes(typedesyscal, 'big')
    # print('Syscal type: {}'.format(syscal_type))

    # comment
    buffer = fid.read(1024)
    comment_raw = struct.iter_unpack('c', buffer)
    comment = ''.join([x[0].decode('utf-8') for x in comment_raw])
    # print('comment', comment)

    metadata = {
        'version': version,
        'syscal_type': syscal_type,
        'comment': comment,
    }

    measurements = []
    # for each measurement
    counter = 0
    while(fid.tell() < total_size):
        dataentry = {}

        # print('COUNTER', counter)
        buffer = fid.read(2)
        array_type = struct.unpack('h', buffer)
        array_type
        # print(array_type)
        # not used
        moretmeasure = fget(fid, 'h', 2)
        moretmeasure
        # print('moretmeasure', moretmeasure)
        # measurement time [ms]
        mtime = fget(fid, 'f', 4)
        # print('measurement time', mtime)
        # delay before IP measurements start [ms]
        mdelay = fget(fid, 'f', 4)
        # print('mdelay', mdelay)
        TypeCpXyz = fget(fid, 'h', 2)
        # our file format documentation always assumes this value to be == 1
        assert TypeCpXyz == 1
        # print('TypeCpXyz', TypeCpXyz)
        # ignore
        fget(fid, 'h', 2)
        # positions: a b m n [m]
        xpos = fget(fid, '4f', 16)
        # print('xpos', xpos)
        ypos = fget(fid, '4f', 16)
        # print('ypos', ypos)
        zpos = fget(fid, '4f', 16)
        # print('zpos', zpos)
        # self-potential [mV]
        sp = fget(fid, 'f', 4)
        # print('sp', sp)
        # measured voltage at voltage electrodes m and n [mV]
        vp = fget(fid, 'f', 4)
        # print('vp', vp)
        Iab = fget(fid, 'f', 4)
        # print('iab', Iab)
        rho = fget(fid, 'f', 4)
        # print('rho', rho)
        m = fget(fid, 'f', 4)
        # print('m', m)
        # standard deviation
        q = fget(fid, 'f', 4)
        # print('q', q)
        # timing windows
        Tm = fget(fid, '20f', 20 * 4)
        Tm = np.array(Tm)
        # print('Tm', Tm)
        # chargeabilities
        Mx = fget(fid, '20f', 20 * 4)
        Mx = np.array(Mx)

        dataentry['version'] = version
        dataentry['mtime'] = mtime
        dataentry['x_a'] = xpos[0]
        dataentry['x_b'] = xpos[1]
        dataentry['x_m'] = xpos[2]
        dataentry['x_n'] = xpos[3]
        dataentry['y_a'] = ypos[0]
        dataentry['y_b'] = ypos[1]
        dataentry['y_m'] = ypos[2]
        dataentry['y_n'] = ypos[3]
        dataentry['z_a'] = zpos[0]
        dataentry['z_b'] = zpos[1]
        dataentry['z_m'] = zpos[2]
        dataentry['z_n'] = zpos[3]
        dataentry['mdelay'] = mdelay
        dataentry['vp'] = vp
        dataentry['q'] = q
        dataentry['m'] = m
        dataentry['Tm'] = Tm
        dataentry['Mx'] = Mx
        dataentry['sp'] = sp
        dataentry['Iab'] = Iab
        dataentry['rho'] = rho

        if moretmeasure > 0:
            # print('Mx', Mx)
            # this is 4 bytes used to store information on the measured channel
            # Channel + NbChannel
            buffer = fid.read(1)
            buffer_bin = bin(ord(buffer))[2:].rjust(8, '0')
            # print(buffer_bin)
            channel = int(buffer_bin[4:], 2)
            channelnb = int(buffer_bin[0:4], 2)
            # print('ChannelNB:', channelnb)
            # print('Channel:', channel)
            # 4 binaries + unused
            buffer = fid.read(1)
            buffer_bin = bin(ord(buffer))[2:].rjust(8, '0')
            # print(buffer_bin)
            overload = bool(int(buffer_bin[4]))
            channel_valid = bool(int(buffer_bin[5]))
            channel_sync = bool(int(buffer_bin[6]))
            gap_filler = bool(int(buffer_bin[7]))
            # print(overload, channel_valid, channel_sync, gap_filler)
            measurement_num = fget(fid, 'H', 2)
            # print('measurement_num', measurement_num)
            filename = fget(fid, '12s', 12)
            # print('filename', filename)
            latitude = fget(fid, 'f', 4)
            # print('lat', latitude)
            longitude = fget(fid, 'f', 4)
            # print('long', longitude)
            # number of stacks
            NbCren = fget(fid, 'f', 4)
            # print('Stacks', NbCren)
            # RS check
            RsChk = fget(fid, 'f', 4)
            # print('RsChk', RsChk)
            dataentry['channel'] = channel
            dataentry['overload'] = overload
            dataentry['channel_valid'] = channel_valid
            dataentry['channel_sync'] = channel_sync
            dataentry['gap_filler'] = gap_filler
            dataentry['NbStacks'] = NbCren
            dataentry['RsCHk'] = RsChk,
            dataentry['latitude'] = latitude
            dataentry['longitude'] = longitude
            dataentry['channelnb'] = channelnb
            dataentry['measurement_num'] = measurement_num
            dataentry['nr'] = measurement_num

            if moretmeasure >= 2:
                # absolute applied voltage
                vab = fget(fid, 'f', 4)
                # print('Vab', vab)
                # tx battery voltage [V]
                batTX = fget(fid, 'f', 4)
                # print('batTX', batTX)
                # rx battery voltage [V]
                batRX = fget(fid, 'f', 4)
                # print('batRX', batRX)
                temperature = fget(fid, 'f', 4)
                # print('Temp.', temperature)
                dataentry['vab'] = vab
                dataentry['batTX'] = batTX
                dataentry['batRX'] = batRX
                dataentry['temperature'] = temperature

            if moretmeasure == 3:
                # TODO: date and time not analyzed
                b = struct.unpack('2f', fid.read(2 * 4))
                # print('b', b)
                b

        measurements.append(dataentry)

        counter += 1

    # create a dataframe with all primary data
    df = pd.DataFrame(
        measurements
    ).reset_index()
    return metadata, df



"""Write CRMod/CRTomo (Kemna, 2000) compatible files
"""
import os
import numpy as np


def save_block_to_crt(filename, group, norrec='all', store_errors=False):
    """Save a dataset to a CRTomo-compatible .crt file

    Parameters
    ----------
    filename : string
        Output filename
    group : pandas.group
        Data group
    norrec : string
        Which data to export Possible values: all|norrec|nor|rec
    store_errors : bool
        If true, store errors of the data in a separate column

    """
    if norrec not in ('all', 'norrec'):
        assert norrec in ('nor', 'rec')
        group = group.query('norrec == "{0}"'.format(norrec))

    # todo: we need to fix the global naming scheme for columns!
    with open(filename, 'wb') as fid:
        fid.write(
            bytes('{0}\n'.format(len(group)), 'UTF-8')
        )

        AB = group['a'] * 1e4 + group['b']
        MN = group['m'] * 1e4 + group['n']
        line = [
            AB.values.astype(int),
            MN.values.astype(int),
            group['r'].values,
        ]

        if 'rpha' in group:
            line.append(group['rpha'].values)
        else:
            line.append(group['r'].values * 0.0)

        fmt = '%i %i %f %f'
        if store_errors:
            line += (
                group['r_error'].values,
                group['rpha_error'].values,
            )
            fmt += ' %f %f'

        subdata = np.array(line).T
        np.savetxt(fid, subdata, fmt=fmt)


def write_files_to_directory(df, directory, **kwargs):
    """Write sEIT data ta files. Data of each frequency is written in a
    separate file that conforms to the CRMod/CRTomo standard, and can directly
    be used for inversions using CRTomo.

    Parameters
    ----------
    df : pandas.DataFrame
        Data
    directory : string
        Output directory. Will be created if not existant

    Other Parameters
    ----------------
    store_errors: bool, optional
        store the device generated errors in the output files (as additional
        columns). Default: False
    norrec: string, optional
        all|nor|rec which normal-reciprocal data set to use. Default: all

    """
    if 'frequency' in df.columns:
        group_key = 'frequency'
    elif 'frequency_[Hz]' in df.columns:
        group_key = 'frequency_[Hz]'
    else:
        group_key = None

    if not os.path.isdir(directory):
        os.makedirs(directory)

    pwd = os.getcwd()
    os.chdir(directory)

    if group_key is not None:
        g = df.groupby(group_key)

        nr = 1
        frequencies_used = []
        for frequency, group in sorted(g):
            if group.shape[0] > 0:
                frequencies_used.append(frequency)
            filename = 'volt_{0:02}_{1:.6}Hz.crt'.format(nr, frequency)
            save_block_to_crt(
                filename,
                group,
                norrec=kwargs.get('norrec', 'all'),
                store_errors=kwargs.get('store_errors', False),
            )

            nr += 1
        np.savetxt('frequencies.dat', frequencies_used)
    else:
        save_block_to_crt(
            'volt.dat',
            df,
            norrec=kwargs.get('norrec', 'all'),
            store_errors=kwargs.get('store_errors', False),
        )

    os.chdir(pwd)



# -*- coding: utf-8 -*-
# medusa data file format: FZJ-EZ-2017
import datetime
import pandas as pd
import numpy as np


def _extract_adc_data(mat, **kwargs):
    """Extract adc-channel related data (i.e., data that is captured for all 48
    channels of the 40-channel medusa system

    Data not imported:

    * Ue3
    * Zs3
    * Us3
    """

    md = mat['MD'].squeeze()
    # Labview epoch
    epoch = datetime.datetime(1904, 1, 1)

    def convert_epoch(x):
        timestamp = epoch + datetime.timedelta(seconds=x.astype(float))
        return timestamp

    dfl = []
    # loop over frequencies
    for f_id in range(0, md.size):
        # print('Frequency: ', emd[f_id]['fm'])
        fdata = md[f_id]
        # for name in fdata.dtype.names:
        #     print(name, fdata[name].shape)

        timestamp = np.atleast_2d(
            [convert_epoch(x) for x in np.atleast_1d(fdata['Time']).flatten()]
        ).T
        data_list = []
        for key, data in zip(
                (
                    'Ug3_1',
                    'Ug3_2',
                    'Ug3_3',
                    'Us3_1',
                    'Us',
                    'Us3_2',
                    'Us3_3',
                    'Ue3_1',
                    'Ue3_2',
                    'Ue3_3',
                ), (
                    fdata['Ug3'][:, :, 0],
                    fdata['Ug3'][:, :, 1],
                    fdata['Ug3'][:, :, 2],
                    np.mean(
                        fdata['Us3'],
                        axis=2,
                    ),
                    fdata['Us3'][:, :, 0],
                    fdata['Us3'][:, :, 1],
                    fdata['Us3'][:, :, 2],
                    fdata['Ue3'][:, :, 0],
                    fdata['Ue3'][:, :, 1],
                    fdata['Ue3'][:, :, 2],
                )):
            df = pd.DataFrame(
                data, columns=['ch{:02}'.format(i) for i in range(48)]).T
            df['parameter'] = key
            df.set_index('parameter', append=True, inplace=True)
            df = df.T
            # I THINK we can just sort a, b - the actual data should already
            # conform to the notation that the lower electrode number is
            # assigned to a
            df[['a', 'b']] = pd.DataFrame(np.sort(fdata['cni'], axis=1))
            # df['b'] = fdata['cni'][:, 1]
            # df['a'] = fdata['cni'][:, 0]
            df.set_index(['a', 'b'], inplace=True)
            data_list.append(df)

        # merge everything
        df_all = data_list[0]
        if len(data_list) > 1:
            for subdata in data_list[1:]:
                df_all = pd.merge(
                    df_all, subdata, left_index=True, right_index=True)

        df_all['datetime'] = timestamp
        df_all['frequency'] = fdata['fm']
        dfl.append(df_all)

    dfl = pd.concat(dfl)
    dfl.set_index('frequency', append=True, inplace=True)
    dfl.sort_index(axis=0, inplace=True)
    dfl.sort_index(axis=1, inplace=True)

    dfl.columns.set_names('channel', level=0, inplace=True)
    dfl.set_index('datetime', append=True, inplace=True)

    return dfl


def _extract_md(mat, **kwargs):
    md = mat['MD'].squeeze()
    # current channel adc numbers:
    # 1: before shunt, first channel
    # 2: second shunt, first channel
    # 3: before shunt, second channel
    # 4: second shunt, second channel
    nai = mat['MP']['NAI'].take(0) - 1

    # Labview epoch
    epoch = datetime.datetime(1904, 1, 1)

    def convert_epoch(x):
        timestamp = epoch + datetime.timedelta(seconds=x.astype(float))
        return timestamp

    dfl = []
    # loop over frequencies
    for f_id in range(0, md.size):
        # print('Frequency: ', emd[f_id]['fm'])
        fdata = md[f_id]
        timestamp = np.atleast_2d(
            [convert_epoch(x) for x in np.atleast_1d(fdata['Time']).flatten()]
        ).T
        """
        From Egon:
        In Zg3 sind jetzt die Werte gegen Masse dargestellt.
        Zg3(:,1,:) ist die Impedanz U(A-Gnd)/IA und
        Zg3(:,2,:) ist die Impedanz U(B-Gnd)/IB
        Mit den Stromelektroden A und B.
        Die Summe von beiden ist die Impedanz Zg3(AB)
        """

        def _to_3d(array):
            if len(array.shape) == 2:
                return array[np.newaxis, :, :]
            return array

        Zg3_complex = np.atleast_2d(
            np.sum(_to_3d(fdata['Zg3']), axis=1)
        )
        df = pd.DataFrame()
        # I THINK we can just sort a, b - the actual data should already
        # conform to the notation that the lower electrode number is assigned
        # to a
        df[['a', 'b']] = pd.DataFrame(
            np.sort(np.atleast_2d(fdata['cni']), axis=1))
        df['datetime'] = timestamp
        df['frequency'] = fdata['fm']
        df['U0'] = fdata['U0']
        df['Zg1_ela'] = _to_3d(fdata['Zg3'])[:, 0, 0]
        df['Zg2_ela'] = _to_3d(fdata['Zg3'])[:, 0, 1]
        df['Zg3_ela'] = _to_3d(fdata['Zg3'])[:, 0, 2]
        df['Zg1_elb'] = _to_3d(fdata['Zg3'])[:, 1, 0]
        df['Zg2_elb'] = _to_3d(fdata['Zg3'])[:, 1, 1]
        df['Zg3_elb'] = _to_3d(fdata['Zg3'])[:, 1, 2]
        df['Zg1'] = Zg3_complex[:, 0]
        df['Zg2'] = Zg3_complex[:, 1]
        df['Zg3'] = Zg3_complex[:, 2]

        df['Zg'] = np.mean(df[['Zg1', 'Zg2', 'Zg3']].values, axis=1)

        df['Cl1'] = np.atleast_2d(fdata['Cl3'])[:, 0]
        df['Cl2'] = np.atleast_2d(fdata['Cl3'])[:, 1]
        df['Cl3'] = np.atleast_2d(fdata['Cl3'])[:, 2]

        # leakage capacitance
        df['Cl'] = np.mean(df[['Cl1', 'Cl2', 'Cl3']].values, axis=1)

        df['Is1'] = np.atleast_2d(fdata['Is3'])[:, 0]
        df['Is2'] = np.atleast_2d(fdata['Is3'])[:, 1]
        df['Is3'] = np.atleast_2d(fdata['Is3'])[:, 2]
        df['Is'] = np.mean(np.atleast_2d(fdata['Is3']), axis=1)
        df['Il1'] = np.atleast_2d(fdata['Il3'])[:, 0]
        df['Il2'] = np.atleast_2d(fdata['Il3'])[:, 1]
        df['Il3'] = np.atleast_2d(fdata['Il3'])[:, 2]
        df['Il'] = np.mean(np.atleast_2d(fdata['Il3']), axis=1)

        # "standard" injected current, in [mA]
        df['Iab'] = np.abs(df['Is']) * 1e3

        # take absolute value and convert to mA
        df['Ileakage'] = np.abs(df['Il']) * 1e3

        df['Ii1_ela'] = _to_3d(fdata['Ii3'])[:, 0, 0]
        df['Ii2_ela'] = _to_3d(fdata['Ii3'])[:, 0, 1]
        df['Ii3_ela'] = _to_3d(fdata['Ii3'])[:, 0, 2]
        df['Ii1_elb'] = _to_3d(fdata['Ii3'])[:, 1, 0]
        df['Ii2_elb'] = _to_3d(fdata['Ii3'])[:, 1, 1]
        df['Ii3_elb'] = _to_3d(fdata['Ii3'])[:, 1, 2]

        df['Uc1'] = np.atleast_2d(fdata['Uc3'])[:, 0]
        df['Uc2'] = np.atleast_2d(fdata['Uc3'])[:, 1]
        df['Uc3'] = np.atleast_2d(fdata['Uc3'])[:, 2]

        df['Yg1_ela'] = _to_3d(fdata['Yg3'])[:, 0, 0]
        df['Yg2_ela'] = _to_3d(fdata['Yg3'])[:, 0, 1]
        df['Yg3_ela'] = _to_3d(fdata['Yg3'])[:, 0, 2]
        df['Yg1_elb'] = _to_3d(fdata['Yg3'])[:, 1, 0]
        df['Yg2_elb'] = _to_3d(fdata['Yg3'])[:, 1, 1]
        df['Yg3_elb'] = _to_3d(fdata['Yg3'])[:, 1, 2]

        df['Cg1_ela'] = _to_3d(fdata['Cg3'])[:, 0, 0]
        df['Cg2_ela'] = _to_3d(fdata['Cg3'])[:, 0, 1]
        df['Cg3_ela'] = _to_3d(fdata['Cg3'])[:, 0, 2]
        df['Cg1_elb'] = _to_3d(fdata['Cg3'])[:, 1, 0]
        df['Cg2_elb'] = _to_3d(fdata['Cg3'])[:, 1, 1]
        df['Cg3_elb'] = _to_3d(fdata['Cg3'])[:, 1, 2]

        df['Yl1'] = np.atleast_2d(fdata['Yl3'])[:, 0]
        df['Yl2'] = np.atleast_2d(fdata['Yl3'])[:, 1]
        df['Yl3'] = np.atleast_2d(fdata['Yl3'])[:, 2]

        df['U1_Shunt1_before'] = _to_3d(fdata['Us3'])[:, nai[0], 0]
        df['U2_Shunt1_before'] = _to_3d(fdata['Us3'])[:, nai[0], 1]
        df['U3_Shunt1_before'] = _to_3d(fdata['Us3'])[:, nai[0], 2]

        df['U1_Shunt1_after'] = _to_3d(fdata['Us3'])[:, nai[1], 0]
        df['U2_Shunt1_after'] = _to_3d(fdata['Us3'])[:, nai[1], 1]
        df['U3_Shunt1_after'] = _to_3d(fdata['Us3'])[:, nai[1], 2]

        df['U1_Shunt2_before'] = _to_3d(fdata['Us3'])[:, nai[2], 0]
        df['U2_Shunt2_before'] = _to_3d(fdata['Us3'])[:, nai[2], 1]
        df['U3_Shunt2_before'] = _to_3d(fdata['Us3'])[:, nai[2], 2]

        df['U1_Shunt2_after'] = _to_3d(fdata['Us3'])[:, nai[3], 0]
        df['U2_Shunt2_after'] = _to_3d(fdata['Us3'])[:, nai[3], 1]
        df['U3_Shunt2_after'] = _to_3d(fdata['Us3'])[:, nai[3], 2]

        dfl.append(df)

    df = pd.concat(dfl).sort_values(['a', 'b', 'frequency'])
    # print('df')
    # import IPython
    # IPython.embed()

    return df


def _extract_emd(mat, **kwargs):
    """Extract the data from the EMD substruct, given a medusa-created MNU0-mat
    file

    Parameters
    ----------

    mat: matlab-imported struct

    """
    emd = mat['EMD'].squeeze()
    # Labview epoch
    epoch = datetime.datetime(1904, 1, 1)

    def convert_epoch(x):
        timestamp = epoch + datetime.timedelta(seconds=x.astype(float))
        return timestamp

    dfl = []
    # loop over frequencies
    for f_id in range(0, emd.size):
        # print('Frequency: ', emd[f_id]['fm'])
        fdata = emd[f_id]
        # some consistency checks
        if len(fdata['nu']) == 2 and fdata['nu'].shape[1] == 2:
            raise Exception('Need MNU0 file, not a quadpole .mat file:')

        timestamp = pd.DataFrame(
            np.atleast_2d(
                [convert_epoch(x) for x in np.atleast_1d(
                    fdata['Time']).flatten()]
            ).T,
            columns=['datetime', ]
        )
        # well, this must be the most complicated way to get the unique
        # injections
        y = np.ascontiguousarray(fdata['ni']).view(
            np.dtype(
                (np.void, fdata['ni'].dtype.itemsize * fdata['ni'].shape[1])
            )
        )
        _, idx = np.unique(y, return_index=True)
        ab = pd.DataFrame(fdata['ni'][np.sort(idx)], columns=['a', 'b'])
        timestamp[['a', 'b']] = ab
        # print(timestamp.shape)
        # print(fdata['ni'].shape)
        # print('nu', fdata['nu'].shape)
        # print(fdata['Zt3'].shape)
        # exit()
        df = pd.DataFrame()
        df[['a', 'b']] = pd.DataFrame(fdata['ni'])
        df = pd.merge(df, timestamp, left_on=['a', 'b'], right_on=['a', 'b'])
        df['p'] = fdata['nu']
        df[['Zt1', 'Zt2', 'Zt3']] = pd.DataFrame(fdata['Zt3'])
        # df[['Zg1', 'Zg2', 'Zg3']] = pd.DataFrame(fdata['Zg3'])
        # df[['Is1', 'Is2', 'Is3']] = pd.DataFrame(fdata['Is3'])
        # df[['Il1', 'Il2', 'Il3']] = pd.DataFrame(fdata['Il3'])

        # import IPython
        # IPython.embed()
        # exit()

        # df = pd.DataFrame(
        #     np.hstack((
        #         # timestamp,
        #         fdata['ni'],
        #         fdata['nu'][:, np.newaxis],
        #         fdata['Zt3'],
        #         # fdata['Is3'],
        #         # fdata['Il3'],
        #         # fdata['Zg3'],
        #         # fdata['As3'][:, 0, :].squeeze(),
        #         # fdata['As3'][:, 1, :].squeeze(),
        #         # fdata['As3'][:, 2, :].squeeze(),
        #         # fdata['As3'][:, 3, :].squeeze(),
        #         # fdata['Yg13'],
        #         # fdata['Yg23'],
        #     )),
        # )
        # df.columns = (
        #     # 'datetime',
        #     'a',
        #     'b',
        #     'p',
        #     # 'Is1',
        #     # 'Is2',
        #     # 'Is3',
        #     # 'Il1',
        #     # 'Il2',
        #     # 'Il3',
        #     # 'Zg1',
        #     # 'Zg2',
        #     # 'Zg3',
        #     # 'ShuntVoltage1_1',
        #     # 'ShuntVoltage1_2',
        #     # 'ShuntVoltage1_3',
        #     # 'ShuntVoltage2_1',
        #     # 'ShuntVoltage2_2',
        #     # 'ShuntVoltage2_3',
        #     # 'ShuntVoltage3_1',
        #     # 'ShuntVoltage3_2',
        #     # 'ShuntVoltage3_3',
        #     # 'ShuntVoltage4_1',
        #     # 'ShuntVoltage4_2',
        #     # 'ShuntVoltage4_3',
        #     # 'Yg13_1',
        #     # 'Yg13_2',
        #     # 'Yg13_3',
        #     # 'Yg23_1',
        #     # 'Yg23_2',
        #     # 'Yg23_3',
        # )

        df['frequency'] = np.ones(df.shape[0]) * fdata['fm']

        # cast to correct type
        # df['datetime'] = pd.to_datetime(df['datetime'])
        # df['a'] = df['a'].astype(int)
        # df['b'] = df['b'].astype(int)
        # df['p'] = df['p'].astype(int)

        df['Zt1'] = df['Zt1'].astype(complex)
        df['Zt2'] = df['Zt2'].astype(complex)
        df['Zt3'] = df['Zt3'].astype(complex)

        # df['Zg1'] = df['Zg1'].astype(complex)
        # df['Zg2'] = df['Zg2'].astype(complex)
        # df['Zg3'] = df['Zg3'].astype(complex)

        # df['Is1'] = df['Is1'].astype(complex)
        # df['Is2'] = df['Is2'].astype(complex)
        # df['Is3'] = df['Is3'].astype(complex)

        # df['Il1'] = df['Il1'].astype(complex)
        # df['Il2'] = df['Il2'].astype(complex)
        # df['Il3'] = df['Il3'].astype(complex)

        # df['ShuntVoltage1_1'] = df['ShuntVoltage1_1'].astype(complex)
        # df['ShuntVoltage1_2'] = df['ShuntVoltage1_2'].astype(complex)
        # df['ShuntVoltage1_3'] = df['ShuntVoltage1_3'].astype(complex)

        # df['ShuntVoltage2_1'] = df['ShuntVoltage2_1'].astype(complex)
        # df['ShuntVoltage2_2'] = df['ShuntVoltage2_2'].astype(complex)
        # df['ShuntVoltage2_3'] = df['ShuntVoltage2_3'].astype(complex)

        # df['ShuntVoltage3_1'] = df['ShuntVoltage3_1'].astype(complex)
        # df['ShuntVoltage3_2'] = df['ShuntVoltage3_2'].astype(complex)
        # df['ShuntVoltage3_3'] = df['ShuntVoltage3_3'].astype(complex)

        # df['ShuntVoltage4_1'] = df['ShuntVoltage4_1'].astype(complex)
        # df['ShuntVoltage4_2'] = df['ShuntVoltage4_2'].astype(complex)
        # df['ShuntVoltage4_3'] = df['ShuntVoltage4_3'].astype(complex)

        dfl.append(df)

    if len(dfl) == 0:
        return None
    df = pd.concat(dfl)

    # average swapped current injections here!
    # TODO

    # sort current injections
    condition = df['a'] > df['b']
    df.loc[condition, ['a', 'b']] = df.loc[condition, ['b', 'a']].values
    # for some reason we lose the integer casting of a and b here
    df['a'] = df['a'].astype(int)
    df['b'] = df['b'].astype(int)
    # change sign because we changed A and B
    df.loc[condition, ['Zt1', 'Zt2', 'Zt3']] *= -1

    # average of Z1-Z3
    df['Zt'] = np.mean(df[['Zt1', 'Zt2', 'Zt3']].values, axis=1)
    # we need to keep the sign of the real part
    sign_re = np.real(df['Zt']) / np.abs(np.real(df['Zt']))
    df['r'] = np.abs(df['Zt']) * sign_re
    # df['Zt_std'] = np.std(df[['Z1', 'Z2', 'Z3']].values, axis=1)

    # df['Is'] = np.mean(df[['Is1', 'Is2', 'Is3']].values, axis=1)
    # df['Il'] = np.mean(df[['Il1', 'Il2', 'Il3']].values, axis=1)
    # df['Zg'] = np.mean(df[['Zg1', 'Zg2', 'Zg3']].values, axis=1)

    # "standard" injected current, in [mA]
    # df['Iab'] = np.abs(df['Is']) * 1e3
    # df['Iab'] = df['Iab'].astype(float)
    # df['Is_std'] = np.std(df[['Is1', 'Is2', 'Is3']].values, axis=1)

    # # take absolute value and convert to mA
    # df['Ileakage'] = np.abs(df['Il']) * 1e3
    # df['Ileakage'] = df['Ileakage'].astype(float)

    return df



"""Time-domain IP (induced polarization) container
"""

import numpy as np
import pandas as pd

import reda.utils.mpl


from reda.containers.BaseContainer import ImportersBase
from reda.containers.BaseContainer import BaseContainer

# import reda.importers.bert as reda_bert_import
import reda.importers.iris_syscal_pro as reda_syscal
import reda.importers.mpt_das1 as reda_mpt

from reda.utils.decorators_and_managers import append_doc_of
from reda.utils.decorators_and_managers import LogDataChanges

plt, mpl = reda.utils.mpl.setup()


class TDIPImporters(ImportersBase):
    """This class provides wrappers for most of the importer functions and is
    meant to be inherited by the TDIP data container.

    See Also
    --------
    Exporters
    """

    @append_doc_of(reda_syscal.import_bin)
    def import_syscal_bin(self, filename, **kwargs):
        """Syscal import

        timestep: int or :class:`datetime.datetime`
            if provided use this value to set the 'timestep' column of the
            produced dataframe. Default: 0

        """
        timestep = kwargs.get('timestep', None)
        if 'timestep' in kwargs:
            del (kwargs['timestep'])
        self.logger.info('IRIS Syscal Pro bin import')
        with LogDataChanges(self, filter_action='import'):
            data, electrodes, topography = reda_syscal.import_bin(
                filename, **kwargs
            )
            if timestep is not None:
                data['timestep'] = timestep
            self._add_to_container(data)
        if kwargs.get('verbose', False):
            print('Summary:')
            self._describe_data(data)

    # JG: ensure that there's chargeability involved?
    # @append_doc_of(reda_bert_import.import_ohm)
    # def import_bert(self, filename, **kwargs):
    #     """BERT .ohm file import"""
    #     timestep = kwargs.get('timestep', None)
    #     if 'timestep' in kwargs:
    #         del (kwargs['timestep'])
    #
    #     self.logger.info('Unified data format (BERT/pyGIMLi) file import')
    #     with LogDataChanges(self, filter_action='import',
    #                         filter_query=os.path.basename(filename)):
    #         data, electrodes, topography = reda_bert_import.import_ohm(
    #             filename, **kwargs)
    #         if timestep is not None:
    #             data['timestep'] = timestep
    #         self._add_to_container(data)
    #         self.electrode_positions = electrodes  # See issue #22
    #     if kwargs.get('verbose', False):
    #         print('Summary:')
    #         self._describe_data(data)

    @append_doc_of(reda_mpt.import_das1_td)
    def import_mpt(self, filename, **kwargs):
        """MPT DAS 1 TD importer

        timestep: int or :class:`datetime.datetime`
            if provided use this value to set the 'timestep' column of the
            produced dataframe. Default: 0

        """
        timestep = kwargs.get('timestep', None)
        if 'timestep' in kwargs:
            del (kwargs['timestep'])
        self.logger.info('MPT DAS-1 import')
        with LogDataChanges(self, filter_action='import'):
            data, electrodes, topography = reda_mpt.import_das1_td(
                filename, **kwargs)
            if timestep is not None:
                data['timestep'] = timestep
            self._add_to_container(data)

        if kwargs.get('verbose', False):
            print('Summary:')
            self._describe_data(data)

    # @functools.wraps(import_bert)
    # def import_pygimli(self, *args, **kargs):
    #     self.import_bert(*args, **kargs)


class TDIP(BaseContainer, TDIPImporters):
    """."""

    def __init__(self, data=None, electrode_positions=None, topography=None):
        """
        Parameters
        ----------
        data : :py:class:`pandas.DataFrame`
            If not None, then the provided DataFrame is assumed to contain
            valid data previously prepared elsewhere. Please refer to the
            documentation for required columns.
        electrode_positions : :py:class:`pandas.DataFrame`
            If set, this is expected to be a DataFrame which contains electrode
            positions with columns: "x", "y", "z".
        topography : :py:class:`pandas.DataFrame`
            If set, this is expected to a DataFrame which contains topography
            information with columns: "x", "y", "z".

        """

        self.setup_logger(__name__)
        self.required_columns = ['a',
                                 'b',
                                 'm',
                                 'n',
                                 'r',
                                 'chargeability']
        self.data = self.check_dataframe(data)
        self.electrode_positions = electrode_positions
        self.topography = topography

    def check_dataframe(self, dataframe):
        """Check the given dataframe for the required type and columns
        """
        if dataframe is None:
            return None

        # is this a DataFrame
        if not isinstance(dataframe, pd.DataFrame):
            raise Exception(
                'The provided dataframe object is not a pandas.DataFrame'
            )

        for column in self.required_columns:
            if column not in dataframe:
                raise Exception('Required column not in dataframe: {0}'.format(
                    column
                ))
        return dataframe

    def plot_decay_curve(self, filename=None, index_nor=None, index_rec=None,
                         nr_id=None, abmn=None,
                         return_fig=False):
        """Plot decay curve

        Input scheme: We recognize three ways to specify the quadrupoles to
        plot (in descending priority):

            1) indices for normal/reciprocal
            2) by specifying the id
            3) by specifying abmn (note that here the exact quadrupole must be
               present. For example, when (1,2,4,3) is requested, (2,1,4,3)
               will not be used).

        Parameters
        ----------
        filename : string, optional
            If given, filename to plot to.

        Returns
        -------
        fig : :class:`matplotlib.Figure`
            Figure object, only returned if return_fig=True

        """
        def get_indices_for_id(norrec_id):
            subquery_nor = self.data.query(
                'id == {} and norrec == "nor"'.format(norrec_id)
            )
            if subquery_nor.shape[0] >= 1:
                indices = [subquery_nor.index.values[0], ]
            else:
                indices = [None, ]

            subquery_rec = self.data.query(
                'id == {} and norrec == "rec"'.format(norrec_id)
            )
            if subquery_rec.shape[0] >= 1:
                indices = indices + [subquery_rec.index.values[0], ]
            else:
                indices = indices + [None, ]
            return indices

        # select data to plot
        # 1: indices
        if index_nor is not None or index_rec is not None:
            indices = [index_nor, index_rec]
        elif nr_id is not None:
            # reset the index
            self.data.reset_index(drop=True, inplace=True)
            indices = get_indices_for_id(nr_id)
        elif abmn is not None:
            subquery = self.data.query(
                'a == {} and b == {} and m == {} and n == {}'.format(*abmn)
            )
            # print(abmn)
            # print('subquery', subquery)
            # import IPython
            # IPython.embed()
            if subquery.shape[0] > 0:
                self.data.reset_index(drop=True, inplace=True)
                indices = get_indices_for_id(subquery['id'].values[0])
            else:
                raise Exception(
                    'configuration not found. Perhaps electrodes were ' +
                    'switched due to negative geometric factors?'
                )
        else:
            raise Exception('No selection method successful!')

        # plot
        fig, axes = plt.subplots(1, 3, figsize=(15 / 2.54, 8 / 2.54))
        labels = ('normal', 'reciprocal')

        # gather data
        data_list = []
        data_normed_list = []
        for nr, index in enumerate(indices):
            if index is not None:
                data = self.data.loc[index, 'Mx']
                data_norm = data / self.data.loc[index, 'Iab']
                data_list.append(data)
                data_normed_list.append(data_norm)
            else:
                data_list.append(None)
                data_normed_list.append(None)

        for nr, index in enumerate(indices):
            if index is None:
                continue
            mdelay = self.data.loc[index, 'mdelay']
            times = self.data.loc[index, 'Tm']
            mtime = mdelay + np.cumsum(times)

            ax = axes[0]
            ax.plot(mtime, data_list[nr], '.-', label=labels[nr])

            ax = axes[1]
            ax.plot(
                mtime, data_normed_list[nr], '.-', label=labels[nr] +
                ' I: {:.1f}mA'.format(self.data.loc[index, 'Iab'])
            )

        ax = axes[2]
        if indices[0] is not None and indices[1] is not None:
            residuals = np.array(data_list[1]) - np.array(data_list[0])
            ax.plot(mtime, residuals, '.-', color='r')
            ax.set_xlabel('time [ms]')
            ax.set_ylabel('residual [mV]')
        else:
            ax.set_visible(False)

        # set labels etc.
        ax = axes[0]
        ax.legend(loc='best', fontsize=6.0)
        ax.set_xlabel('time [ms]')
        ax.set_ylabel('m [mV/V]')

        ax = axes[1]
        ax.set_ylabel('normalized decay curve [-]', fontsize=7.0)
        ax.set_title(r'normed on current', fontsize=6.0)
        ax.legend(loc='best', fontsize=6.0)

        fig.tight_layout()
        if filename is not None:
            fig.savefig(filename, dpi=300)

        if return_fig:
            return fig
        else:
            fig.clf()
            plt.close(fig)

    def to_cr(self,
              b=-1.5,
              estimate_b=False,
              conversion_parameters={
                  "phi_min": -100,
                  "phi_max": -1,
                  "n": 100,
                  "nhc": 4,
                  "tpuls": 2,
                  "tmin": 0.12,
                  "tmax": 1.7
              }):

        """Convert container to a complex resistivity container, using the
        CPA-conversion.

        Kemna, 2000

        COMPLEX RESISTIVITY COPPER MlNERALlZATlONt SPECTRA OF PORPHYRY
        Van Voorhis, G. D.; Nelson, P. H.; Drake, T. L.
        Geophysics (1973 Jan 1) 38 (1): 49-60.

        Application of complex resistivity tomography to field data from
        a kerosene-contaminated siteGold Open Access Authors: A. Kemna, E.
        Räkers and A. Binley DOI: 10.3997/2214-4609.201407300

        Gianluca Fiandaca, Esben Auken, Anders Vest Christiansen, and
        Aurélie Gazoty (2012). ”Time-domain-induced polarization:
        Full-decay forward modeling and 1D laterally constrained inversion
        of Cole-Cole parameters.” GEOPHYSICS, 77(3), E213-E225.
        https://doi.org/10.1190/geo2011-0217.1

        Parameters
        ----------

        b : float
            Factor for the conversion of chargeability into phase:
            phase = b * chargeability
            Default value equals -1.5, which is a good approximation for many
            cases

        estimate_b : bool
            If "True" the factor b is estimated according to the CPA model, for
            a given set of conversion parameters (provided via
            "conversion_paramaters")

        conversion_parameters : dict
            Holds the conversion parameters which are the basis for the
            estimation of the factor b. Parameters in the dictionary:

            phi_min : float
                Lower phase boundary
            phi_max : float
                Upper phase boundary
            n : int
                Number of values in the phase field
            nhc : int
                Number of half cycles
            tpuls : float
                Duration of one puls
            tmin : float
                Lower time boundary
            tmax : float
                Upper time boundary

        """

        if estimate_b:

            phi_min = conversion_parameters["phi_min"]
            phi_max = conversion_parameters["phi_max"]
            n = conversion_parameters["n"]
            nhc = conversion_parameters["nhc"] * 2
            tpuls = conversion_parameters["tpuls"]
            tmax = conversion_parameters["tmax"]
            tmin = conversion_parameters["tmin"]

            phi = np.linspace(phi_min, phi_max, n)

            b = -2 * phi / np.pi / 1e3

            v0 = 0
            for k in range(0, nhc):
                for j in range(0, k):
                    temp = ((1 * j) * tpuls) ** b - ((2 * j + 1) * tpuls) ** b
                    v0 = v0 + ((-1)**(j + 1) * temp)

            ztemp = 0
            for k in range(0, nhc):
                for j in range(0, k):
                    temp = (
                        (tmax + (2 * j) * tpuls) ** (b + 1)
                        - (tmax + (2 * j + 1) * tpuls) ** (b + 1)
                        - (tmin + (2 * j) * tpuls) ** (b + 1)
                        + (tmin + (2 * j + 1) * tpuls) ** (b + 1))
                    ztemp = ztemp + ((-1) ** (j + 1)) * temp
            ztemp = ztemp / (b + 1) / (tmax - tmin)
            m = 1e3 * ztemp / v0

            # fit line through m and phi
            # phi = afit + bfit * m

            A = np.ones((n, 2))
            A[:, 0] = m

            bfit, afit = np.linalg.lstsq(A, phi, rcond=None)[0]

            b = bfit

            print("Estimated conversion factor b: {}".format(b))

        else:
            b = b

        data_new = self.data.copy()
        data_new['rpha'] = b * data_new['chargeability']
        # now that we have magnitude and phase, compute the impedance Zt
        data_new['Zt'] = data_new['r'] * np.exp(data_new['rpha'] * 1j / 1000.0)
        cr = reda.CR(data=data_new)
        return cr



"""Container for Spectral Induced Polarization (SIP) measurements
"""
import numpy as np
import pandas as pd

import reda.utils.mpl

from reda.containers.BaseContainer import ImportersBase
from reda.containers.BaseContainer import BaseContainer

import reda.importers.sip04 as reda_sip04
import reda.importers.mpt_das1 as reda_mpt
# from reda.importers.crtomo import load_mod_file

from reda.utils.decorators_and_managers import append_doc_of
from reda.utils.decorators_and_managers import LogDataChanges

plt, mpl = reda.utils.mpl.setup()


class SIPImporters(ImportersBase):
    """This class provides wrappers for most of the importer functions, and is
    meant to be inherited by the data containers
    """

    # @append_doc_of(reda_sip04.import_sip04)
    def import_sip04(self, filename, timestep=None):
        """SIP04 data import

        Parameters
        ----------
        filename: string
            Path to .mat or .csv file containing SIP-04 measurement results

        Examples
        --------

        ::

            import tempfile
            import reda
            with tempfile.TemporaryDirectory() as fid:
                reda.data.download_data('sip04_fs_01', fid)
                sip = reda.SIP()
                sip.import_sip04(fid + '/sip_dataA.mat')

        """
        df = reda_sip04.import_sip04_data(filename)
        if timestep is not None:
            print('adding timestep')
            df['timestep'] = timestep

        self._add_to_container(df)
        print('Summary:')
        self._describe_data(df)

    @append_doc_of(reda_mpt.import_das1_td)
    def import_mpt(self, filename, **kwargs):
        """MPT DAS 1 FD importer

        timestep: int or :class:`datetime.datetime`
            if provided use this value to set the 'timestep' column of the
            produced dataframe. Default: 0

        """
        timestep = kwargs.get('timestep', None)
        if 'timestep' in kwargs:
            del (kwargs['timestep'])
        self.logger.info('MPT DAS-1 import')
        with LogDataChanges(self, filter_action='import'):
            data, electrodes, topography = reda_mpt.import_das1_sip(
                filename, **kwargs)
            if timestep is not None:
                data['timestep'] = timestep
            self._add_to_container(data)

        if kwargs.get('verbose', False):
            print('Summary:')
            self._describe_data(data)


class SIP(BaseContainer, SIPImporters):
    """."""

    def __init__(self, data=None, electrode_positions=None, topography=None):
        """."""
        self.setup_logger()
        self.data = self.check_dataframe(data)
        self.required_columns = [
            'a',
            'b',
            'm',
            'n',
            'r'
            'frequency',
            'rpha'
            'Zt',
        ]
        self.plot_columns = [
            'frequency',
            'Zt'
        ]
        self.electrode_positions = electrode_positions
        self.topography = topography

    def check_dataframe(self, dataframe):
        """Check the given dataframe for the required type and columns
        """
        if dataframe is None:
            return None

        # is this a DataFrame
        if not isinstance(dataframe, pd.DataFrame):
            raise Exception(
                'The provided dataframe object is not a pandas.DataFrame'
            )

        for column in self.required_columns:
            if column not in dataframe:
                raise Exception('Required column not in dataframe: {0}'.format(
                    column
                ))
        return dataframe

    def reduce_duplicate_frequencies(self):
        """In case multiple frequencies were measured, average them and compute
        std, min, max values for zt.

        In case timesteps were added (i.e., multiple separate measurements),
        group over those and average for each timestep.

        Examples
        --------

        ::

            import tempfile
            import reda
            with tempfile.TemporaryDirectory() as fid:
                reda.data.download_data('sip04_fs_06', fid)
                sip = reda.SIP()
                sip.import_sip04(fid + '/sip_dataA.mat', timestep=0)
                # well, add the spectrum again as another timestep
                sip.import_sip04(fid + '/sip_dataA.mat', timestep=1)
            df = sip.reduce_duplicate_frequencies()

        """
        group_keys = ['frequency', ]
        if 'timestep' in self.data.columns:
            group_keys = group_keys + ['timestep', ]

        g = self.data.groupby(group_keys)

        def group_apply(item):
            y = item[['zt_1', 'zt_2', 'zt_3']].values.flatten()
            zt_imag_std = np.std(np.imag(y))
            zt_real_std = np.std(np.real(y))
            zt_imag_min = np.min(np.imag(y))
            zt_real_min = np.min(np.real(y))
            zt_imag_max = np.max(np.imag(y))
            zt_real_max = np.max(np.real(y))
            zt_imag_mean = np.mean(np.imag(y))
            zt_real_mean = np.mean(np.real(y))
            dfn = pd.DataFrame(
                {
                    'zt_real_mean': zt_real_mean,
                    'zt_real_std': zt_real_std,
                    'zt_real_min': zt_real_min,
                    'zt_real_max': zt_real_max,
                    'zt_imag_mean': zt_imag_mean,
                    'zt_imag_std': zt_imag_std,
                    'zt_imag_min': zt_imag_min,
                    'zt_imag_max': zt_imag_max,
                },
                index=[0, ]
            )

            dfn['count'] = len(y)
            dfn.index.name = 'index'
            return dfn

        p = g.apply(group_apply)
        p.index = p.index.droplevel('index')
        if len(group_keys) > 1:
            p = p.swaplevel(0, 1).sort_index()
        return p

    def frequencies(self):
        """Return all frequencies"""
        return np.array(list(self.data.groupby('frequency').groups.keys()))

    def export_specs_to_ascii(self, frequency_file, data_file):
        """

        """
        subdata = pd.pivot_table(
            self.data.sort_values('frequency'),
            index=['a', 'b', 'm', 'n'],
            columns='frequency',
            values=['r', 'rpha']
        )
        np.savetxt(frequency_file, self.frequencies())
        np.savetxt(data_file, subdata.values)



"""Exporter functions for IRIS Instruments Syscal Pro
"""


def _syscal_write_electrode_coords(fid, spacing, N):
    """helper function that writes out electrode positions to a file descriptor

    Parameters
    ----------
    fid: file descriptor
        data is written here
    spacing: float
        spacing of electrodes
    N: int
        number of electrodes
    """
    fid.write('# X Y Z\n')
    for i in range(0, N):
        fid.write('{0} {1} {2} {3}\n'.format(i + 1, i * spacing, 0, 0))


def _syscal_write_quadpoles(fid, quadpoles):
    """helper function that writes the actual measurement configurations to a
    file descriptor.

    Parameters
    ----------
    fid: file descriptor
        data is written here
    quadpoles: numpy.ndarray
        measurement configurations

    """
    fid.write('# A B M N\n')
    for nr, quadpole in enumerate(quadpoles):
        fid.write(
            '{0} {1} {2} {3} {4}\n'.format(
                nr, quadpole[0], quadpole[1], quadpole[2], quadpole[3]))


def syscal_save_to_config_txt(filename, configs, spacing=1):
    """Write configurations to a Syscal ascii file that can be read by the
    Electre Pro program.

    Parameters
    ----------
    filename: string
        output filename
    configs: numpy.ndarray
        Nx4 array with measurement configurations A-B-M-N

    """
    print('Number of measurements: ', configs.shape[0])
    number_of_electrodes = configs.max().astype(int)

    with open(filename, 'w') as fid:
        _syscal_write_electrode_coords(fid, spacing, number_of_electrodes)
        _syscal_write_quadpoles(fid, configs.astype(int))



#!/usr/bin/env python
import numpy as np
import pandas as pd

import reda


def _get_dataframe():
    """Return a suitable test dataframe"""
    df = pd.DataFrame(
            [
                (0, 1, 2, 4, 3, 1.1, 11.1, 99.8 - 4.99j),
                (0, 1, 2, 5, 4, 1.2, 12.2, 99.7 - 4.4j),
                (0, 1, 2, 6, 5, 1.3, 13.2, 99.7 - 4.2j),
                (0, 1, 2, 7, 6, 1.4, 14.3, 99.7 - 4.1j),
                (0, 2, 3, 5, 4, 1.5, 14.7, 99.7 - 4.998j),
                (0, 2, 3, 6, 5, 1.6, 15.3, 99.7 - 4j),
                (0, 2, 3, 7, 6, 1.7, 16.3, 99.7 - 4.4j),
                (0, 3, 4, 6, 5, 1.8, 17.3, 99.7 - 4.2j),
                (0, 3, 4, 7, 6, 1.9, 18.3, 99.2 - 4j),
                (0, 4, 5, 7, 6, 2.0, 19.3, 99.1 - 3j),
            ],
            columns=[
                'timestep', 'a', 'b', 'm', 'n', 'r', 'Vmn', 'Zt',
            ],
        )
    df['rpha'] = np.arctan2(np.imag(df['Zt']), np.real(df['Zt'])) * 1000
    return df


def test_neg_K():
    """Test computation and subsequent sign correction for two negative K
    factor """
    df = _get_dataframe()

    # test for 1-3 rows that need fixing (i.e., negative K factor + neg. r)
    for rows_to_changes in ([0], [0, 1], [0, 1, 2]):
        # witch m, n of first configuration to yield a negative k factor, r,
        # and rho_a value
        df.iloc[
            rows_to_changes, [3, 4]
        ] = df.iloc[rows_to_changes, [4, 3]].values
        df.loc[df.index[rows_to_changes], ['r', 'Vmn', 'Zt']] *= -1

        ert = reda.ERT(data=df)
        ert.compute_K_analytical(spacing=1, debug=True)

        assert np.all(
            ert.data.loc[
                ert.data.index[rows_to_changes], ['r', 'rho_a', 'k', 'Vmn', ]
            ] >= 0
        )
        # check real and imaginary parts pf Zt
        assert np.all(np.real(ert.data.loc[
            ert.data.index[rows_to_changes], ['Zt']
        ]) >= 0)
        assert np.all(np.imag(ert.data.loc[
            ert.data.index[rows_to_changes], ['Zt']
        ]) <= 0)


if __name__ == '__main__':
    test_neg_K()



# define various labels/units for common SIP parameters

labels = {
    'rmag': {
        'material': {
            'latex': r'$|\rho|~[\Omega m]$',
            'mathml': r'$|\rho| [\Omega m]$',
        },
        'meas': {
            'latex': r'$|Z|~[\Omega]$',
            'mathml': r'$|Z| [\Omega]$',
        },
    },
    'rpha': {
        'material': {
            'latex': r'$-\phi~[mrad]$',
            'mathml': r'$-\phi [mrad]$',
        },
        'meas': {
            'latex': r'$-\varphi~[mrad]$',
            'mathml': r'$-\varphi [mrad]$',
        },
    },
    'cre': {
        'material': {
            'latex': r"$\sigma'~[S/m]$",
            'mathml': r"$\sigma' [S/m]$",
        },
        'meas': {
            'latex': r"$Y'~[S]$",
            'mathml': r"$Y' [S]$",
        },
    },
    'cim': {
        'material': {
            'latex': r"$\sigma''~[S/m]$",
            'mathml': r"$\sigma'' [S/m]$",
        },
        'meas': {
            'latex': r"$Y''~[S]$",
            'mathml': r"$Y'' [S]$",
        },
    },
}


def get_label(parameter, ptype, flavor=None, mpl=None):
    """Return the label of a given SIP parameter

    Parameters
    ----------
    parameter : str
        type of parameter, e.g. rmag|rpha|cre|cim
    ptype : string
        material|meas. Either return the material property (e.g. resistivity)
        or the measurement parameter (e.g., impedance)
    flavor : string, optional
        if set, must be one of latex|mathml. Return a label for latex
        processing, or for mathml processing
    mpl : matplotlib, optional
        if set, infer flavor from mpl.rcParams. Will not be used if flavor is
        set

    Returns
    -------
    label : string
        the requested label
    """
    # determine flavor
    if flavor is not None:
        if flavor not in ('latex', 'mathml'):
            raise Exception('flavor not recognized: {}'.format(flavor))
    else:
        if mpl is None:
            raise Exception('either the flavor or mpl must be provided')
        rendering = mpl.rcParams['text.usetex']
        if rendering:
            flavor = 'latex'
        else:
            flavor = 'mathml'

    # check if the requested label is present
    if parameter not in labels:
        raise Exception('parameter not known')
    if ptype not in labels[parameter]:
        raise Exception('ptype not known')
    if flavor not in labels[parameter][ptype]:
        raise Exception('flavor not known')

    return labels[parameter][ptype][flavor]



"""Histogram functions for raw data."""

import pandas as pd
import reda.utils.mpl
# import pylab as plt
# import matplotlib as mpl
# mpl.rcParams['font.size'] = 8.0
import numpy as np

import reda.main.units as units
plt, mpl = reda.utils.mpl.setup()


def _get_nr_bins(count):
    """depending on the number of data points, compute a best guess for an
    optimal number of bins

    https://en.wikipedia.org/wiki/Histogram#Number_of_bins_and_width
    """
    if count <= 30:
        # use the square-root choice, used by Excel and Co
        k = np.ceil(np.sqrt(count))
    else:
        # use Sturges' formula
        k = np.ceil(np.log2(count)) + 1
    return int(k)


def plot_histograms(ertobj, keys, **kwargs):
    """Generate histograms for one or more keys in the given container.

    Parameters
    ----------
    ertobj : container instance or :class:`pandas.DataFrame`
        data object which contains the data.
    keys : str or list of strings
        which keys (column names) to plot
    merge : bool, optional
        if True, then generate only one figure with all key-plots as columns
        (default True)
    log10plot : bool, optional
        default: True
    extra_dims : list, optional
        ?
    nr_bins : None|int
        if an int is given, use this as the number of bins, otherwise use a
        heuristic.

    Examples
    --------
    >>> from reda.plotters import plot_histograms
    >>> from reda.testing import ERTContainer
    >>> figs_dict = plot_histograms(ERTContainer, "r", merge=False)
    Generating histogram plot for key: r

    Returns
    -------
    figures : dict
        dictionary with the generated histogram figures
    """
    # you can either provide a DataFrame or an ERT object
    if isinstance(ertobj, pd.DataFrame):
        df = ertobj
    else:
        df = ertobj.data

    if df.shape[0] == 0:
        raise Exception('No data present, cannot plot')

    if isinstance(keys, str):
        keys = [keys, ]

    figures = {}
    merge_figs = kwargs.get('merge', True)
    if merge_figs:
        nr_x = 2
        nr_y = len(keys)
        size_x = 15 / 2.54
        size_y = 5 * nr_y / 2.54
        fig, axes_all = plt.subplots(nr_y, nr_x, figsize=(size_x, size_y))
        axes_all = np.atleast_2d(axes_all)

    for row_nr, key in enumerate(keys):
        print('Generating histogram plot for key: {0}'.format(key))
        subdata_raw = df[key].values
        subdata = subdata_raw[~np.isnan(subdata_raw)]
        subdata = subdata[np.isfinite(subdata)]

        nr_of_bins = kwargs.get('nr_of_bins', _get_nr_bins(subdata.size))

        subdata_log10_with_nan = np.log10(subdata[subdata > 0])
        subdata_log10 = subdata_log10_with_nan[~np.isnan(
            subdata_log10_with_nan)
        ]

        subdata_log10 = subdata_log10[np.isfinite(subdata_log10)]

        if merge_figs:
            axes = axes_all[row_nr].squeeze()
        else:
            fig, axes = plt.subplots(1, 2, figsize=(10 / 2.54, 5 / 2.54))

        label = units.get_label(key)
        if not mpl.rcParams['text.usetex']:
            label = label.replace('_', '-')
        ax = axes[0]
        ax.hist(
            subdata,
            nr_of_bins,
        )
        ax.set_xlabel(
            label
        )
        ax.set_ylabel('count')
        ax.xaxis.set_major_locator(mpl.ticker.MaxNLocator(5))
        ax.tick_params(axis='both', which='major', labelsize=6)
        ax.tick_params(axis='both', which='minor', labelsize=6)

        if subdata_log10.size > 0:
            ax = axes[1]
            ax.hist(
                subdata_log10,
                nr_of_bins,
            )
            ax.set_xlabel(r'$log_{10}($' + label + ')')
            ax.set_ylabel('count')
            ax.xaxis.set_major_locator(mpl.ticker.MaxNLocator(5))
        else:
            pass
            # del(axes[1])

        fig.tight_layout()

        if not merge_figs:
            figures[key] = fig

    if merge_figs:
        figures['all'] = fig
    return figures


def plot_histograms_extra_dims(dataobj, keys, primary_dim=None, **kwargs):
    """Produce histograms grouped by one extra dimensions. Produce additional
    figures for all extra dimensions.

    The dimension to spread out along subplots is called the primary extra
    dimension.

    Extra dimensions are:

    * timesteps
    * frequency

    If only "timesteps" are present, timesteps will be plotted as subplots.

    If only "frequencies" are present, frequencies will be plotted as subplots.

    If more than one extra dimensions is present, multiple figures will be
    generated.

    Test cases:

    * not extra dims present (primary_dim=None, timestep, frequency)
    * timestep (primary_dim=None, timestep, frequency)
    * frequency (primary_dim=None, timestep, frequency)
    * timestep + frequency (primary_dim=None, timestep, frequency)

    check nr of returned figures.

    Parameters
    ----------
    dataobj : :py:class:`pandas.DataFrame` or reda container
        The data container/data frame which holds the data
    keys :   string|list|tuple|iterable
        The keys (columns) of the dataobj to plot
    primary_dim : string
        primary extra dimension to plot along subplots. If None, the first
        extra dimension found in the data set is used, in the following order:
        timestep, frequency.
    subquery : str, optional
        if provided, apply this query statement to the data before plotting
    log10 : bool
        Plot only log10 transformation of data (default: False)
    lin_and_log10 : bool
        Plot both linear and log10 representation of data (default: False)
    Nx : int, optional
        Number of subplots in x direction

    Returns
    -------
    dim_name : str
        name of secondary dimensions, i.e. the dimensions for which separate
        figures were created
    figures : dict
        dict containing the generated figures. The keys of the dict correspond
        to the secondary dimension grouping keys

    Examples
    --------
    >>> import reda.testing.containers
    >>> ert = reda.testing.containers.ERTContainer_nr
    >>> import reda.plotters.histograms as RH
    >>> dim_name, fig = RH.plot_histograms_extra_dims(ert, ['r', ])

    >>> import reda.testing.containers
    >>> ert = reda.testing.containers.ERTContainer_nr
    >>> import reda.plotters.histograms as RH
    >>> dim_name, fig = RH.plot_histograms_extra_dims(ert, ['r', 'a'])
    """
    if isinstance(dataobj, pd.DataFrame):
        df_raw = dataobj
    else:
        df_raw = dataobj.data

    if kwargs.get('subquery', False):
        df = df_raw.query(kwargs.get('subquery'))
    else:
        df = df_raw

    # define some labels
    edim_labels = {
        'timestep': (
            'time',
            ''
        ),
        'frequency': (
            'frequency',
            'Hz',
        ),
    }

    # prepare data columns to plot
    if isinstance(keys, str):
        keys = [keys, ]
    columns = keys
    N_c = len(columns)

    # create log10 plots?
    if kwargs.get('lin_and_log10', False):
        transformers = ['lin', 'log10']
        N_trafo = 2
    elif kwargs.get('log10', False):
        transformers = ['log10', ]
        N_trafo = 1
    else:
        transformers = ['lin', ]
        N_trafo = 1

    # available extra dimensions
    extra_dims = ('timestep', 'frequency')

    # select dimension to plot into subplots
    if primary_dim is None or primary_dim not in df.columns:
        for extra_dim in extra_dims:
            if extra_dim in df.columns:
                primary_dim = extra_dim
    # now primary_dim is either None (i.e., we don't have any extra dims to
    # group), or it contains a valid column to group

    # now prepare the secondary dimensions for which we create extra figures
    secondary_dimensions = []
    for extra_dim in extra_dims:
        if extra_dim in df.columns and extra_dim != primary_dim:
            secondary_dimensions.append(extra_dim)

    # group secondary dimensions, or create a tuple with all the data
    if secondary_dimensions:
        group_secondary = df.groupby(secondary_dimensions)
    else:
        group_secondary = (('all', df), )

    figures = {}
    for sec_g_name, sec_g in group_secondary:
        # group over primary dimension
        if primary_dim is None:
            group_primary = (('all', sec_g), )
            N_primary = 1
        else:
            group_primary = sec_g.groupby(primary_dim)
            N_primary = group_primary.ngroups

        # determine layout of Figure
        Nx_max = kwargs.get('Nx', 4)
        N = N_primary * N_c * N_trafo
        Nx = min(Nx_max, N)
        Ny = int(np.ceil(N / Nx))

        size_x = 5 * Nx / 2.54
        size_y = 5 * Ny / 2.54
        fig, axes = plt.subplots(
            Ny, Nx,
            figsize=(size_x, size_y),
            sharex=True,
            sharey=True
        )
        axes = np.atleast_2d(axes)

        index = 0
        for p_name, pgroup in group_primary:
            for column in columns:
                for transformer in transformers:
                    # print('{0}-{1}-{2}'.format(ts_name, column, transformer))

                    subdata_raw = pgroup[column].values
                    subdata = subdata_raw[~np.isnan(subdata_raw)]
                    subdata = subdata[np.isfinite(subdata)]

                    if transformer == 'log10':
                        subdata_log10_with_nan = np.log10(subdata[subdata > 0])
                        subdata_log10 = subdata_log10_with_nan[~np.isnan(
                            subdata_log10_with_nan)
                        ]

                        subdata_log10 = subdata_log10[
                            np.isfinite(subdata_log10)
                        ]
                        subdata = subdata_log10

                    ax = axes.flat[index]
                    ax.hist(
                        subdata,
                        _get_nr_bins(subdata.size),
                    )
                    ax.set_xlabel(
                        units.get_label(column)
                    )
                    ax.set_ylabel('count')
                    ax.xaxis.set_major_locator(mpl.ticker.MaxNLocator(3))
                    ax.tick_params(axis='both', which='major', labelsize=6)
                    ax.tick_params(axis='both', which='minor', labelsize=6)
                    # depending on the type of pname, change the format
                    p_name_fmt = '{}'
                    from numbers import Number
                    if isinstance(p_name, Number):
                        p_name_fmt = '{:.4f}'
                    title_str = '{}: ' + p_name_fmt + ' {}'
                    ax.set_title(
                        title_str.format(
                            edim_labels[primary_dim][0],
                            p_name,
                            edim_labels[primary_dim][1],
                        ),
                        fontsize=7.0,
                    )

                    index += 1

        # remove some labels
        for ax in axes[:, 1:].flat:
            ax.set_ylabel('')
        for ax in axes[:-1, :].flat:
            ax.set_xlabel('')

        fig.tight_layout()
        for ax in axes.flat[index:]:
            ax.set_visible(False)
        figures[sec_g_name] = fig
    return '_'.join(secondary_dimensions), figures



# -*- coding: utf8 -*-
"""

"""
import os
from urllib import request
import zipfile

import pandas as pd
# if this is set to a valid directory path, try to fetch data from here
# The path could be a local copy of the data repository
use_local_data_repository = None

repository_url = ''.join((
    'https://raw.githubusercontent.com/geophysics-ubonn/'
    'reda_examples_mw/master/'
))
inventory_filename = 'inventory.dat'


def download_data(identifier, outdir):
    """Download data from a separate data repository for testing.

    Parameters
    ----------
    identifier: string
        The identifier used to find the data set
    outdir: string
        unzip the data in this directory
    """
    # determine target
    if use_local_data_repository is not None:
        url_base = 'file:' + request.pathname2url(
            use_local_data_repository + os.sep)
    else:
        url_base = repository_url

    print('url_base: {}'.format(url_base))
    url = url_base + inventory_filename
    # download inventory file
    filename, headers =request.urlretrieve(url)

    df = pd.read_csv(
        filename,
        delim_whitespace=True,
        comment='#',
        header=None,
        names=['identifier', 'rel_path'],
    )

    # find relative path to data file
    rel_path_query = df.query('identifier == "{}"'.format(identifier))
    if rel_path_query.shape[0] == 0:
        raise Exception('identifier not found')
    rel_path = rel_path_query['rel_path'].values[0]

    # download the file
    url = url_base + rel_path
    print('data url: {}'.format(url))
    filename, headers =request.urlretrieve(url)

    if not os.path.isdir(outdir):
        os.makedirs(outdir)

    zip_obj = zipfile.ZipFile(filename)
    zip_obj.extractall(outdir)
